"""
Sets of configuration values for a GA4GH server.
Any of these values may be overridden by a file on the path
given by the environment variable GA4GH_CONFIGURATION.
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime


class BaseConfig(object):
    """
    Simplest default server configuration.
    """
    MAX_CONTENT_LENGTH = 2 * 1024 * 1024  # 2MB
    MAX_RESPONSE_LENGTH = 1024 * 1024  # 1MB
    REQUEST_VALIDATION = True
    DEFAULT_PAGE_SIZE = 100
    DATA_SOURCE = "empty://"

    # Options for the simulated backend.
    SIMULATED_BACKEND_RANDOM_SEED = 0
    SIMULATED_BACKEND_NUM_CALLS = 1
    SIMULATED_BACKEND_VARIANT_DENSITY = 0.5
    SIMULATED_BACKEND_NUM_VARIANT_SETS = 1
    SIMULATED_BACKEND_NUM_REFERENCE_SETS = 1
    SIMULATED_BACKEND_NUM_REFERENCES_PER_REFERENCE_SET = 1
    SIMULATED_BACKEND_NUM_ALIGNMENTS_PER_READ_GROUP = 2
    SIMULATED_BACKEND_NUM_READ_GROUPS_PER_READ_GROUP_SET = 2
    SIMULATED_BACKEND_NUM_PHENOTYPE_ASSOCIATIONS = 2
    SIMULATED_BACKEND_NUM_PHENOTYPE_ASSOCIATION_SETS = 2
    SIMULATED_BACKEND_NUM_RNA_QUANTIFICATION_SETS = 2
    SIMULATED_BACKEND_NUM_EXPRESSION_LEVELS_PER_RNA_QUANT_SET = 2

    FILE_HANDLE_CACHE_MAX_SIZE = 50

    LANDING_MESSAGE_HTML = "landing_message.html"
    INITIAL_PEERS = "ga4gh/server/templates/initial_peers.txt"


class ComplianceConfig(BaseConfig):
    """
    Configuration used for compliance testing.
    """
    DATA_SOURCE = "ga4gh-compliance-data/registry.db"
    DEBUG = True


class DevelopmentConfig(BaseConfig):
    """
    Configuration used for development.
    """
    DATA_SOURCE = "ga4gh-example-data/registry.db"
    DEBUG = True


class LocalOidConfig(DevelopmentConfig):
    """
    Configuration used for developing against a local OIDC server
    """
    SECRET_KEY = "super_secret"
    OIDC_PROVIDER = "https://localhost:8443"


class SimulatedConfig(BaseConfig):
    """
    A configuration that uses simulated backing for testing.
    """
    DATA_SOURCE = "simulated://"
    DEBUG = True
    REQUEST_VALIDATION = True


class ProductionConfig(BaseConfig):
    """
    Configuration that is a good basis for production deployments.
    """
    REQUEST_VALIDATION = True
    # We should complain loudly if data source is not set, rather than
    # mysteriously serve no data.
    DATA_SOURCE = None


class GoogleOidcConfig(ProductionConfig):
    """
    Configuration that is a good basis for production deployments using
    Google as the authentication provider.
    """
    SECRET_KEY = "super_secret"
    OIDC_PROVIDER = "https://accounts.google.com"
    OIDC_REDIRECT_URI = "https://localhost/oauth2callback"
    OIDC_CLIENT_ID = "XXX"
    OIDC_CLIENT_SECRET = "XXX"


class TestAuth0Config(DevelopmentConfig):
    """
    Auth0 configuration
    """
    AUTH0_ENABLED = True
    SECRET_KEY = "super_secret"
    AUTH0_SCOPES = "openid email"
    AUTH0_CALLBACK_URL = "http://localhost:8000/callback"
    AUTH0_HOST = "david4096.auth0.com"
    AUTH0_CLIENT_ID = "r99hdj5hhkazgePB5oMYK9Sv4NaUwwYp"
    AUTH0_CLIENT_SECRET = \
        "KeV2tMyGaSgLeOhpoGs_XLH65Tfw43yBjT8DIpaTxXAKmd_bguJwXA6T7D0iYfgB"
    AUTH0_AUTHORIZED_EMAILS = "davidcs@ucsc.edu,your@email.com"


class TestConfig(BaseConfig):
    """
    Configuration used in frontend unit tests.
    """
    TESTING = True
    REQUEST_VALIDATION = True


class TestOidcConfig(TestConfig):
    SECRET_KEY = "super_secret"
    OIDC_PROVIDER = "https://accounts.example.com"
    OIDC_CLIENT_ID = "XXX"
    OIDC_CLIENT_SECRET = "XXX"
    OIDC_AUTHZ_ENDPOINT = "https://accounts.example.com/auth"
    OIDC_TOKEN_ENDPOINT = "https://accounts.example.com/token"
    OIDC_TOKEN_REV_ENDPOINT = "https://accounts.example.com/revoke"


class FlaskDefaultConfig(object):
    """
    The default values for the Flask config.
    Only used in testing.
    """
    APPLICATION_ROOT = None
    DEBUG = False
    JSONIFY_PRETTYPRINT_REGULAR = True
    JSON_AS_ASCII = True
    JSON_SORT_KEYS = True
    LOGGER_NAME = 'ga4gh.frontend'
    MAX_CONTENT_LENGTH = None
    PERMANENT_SESSION_LIFETIME = datetime.timedelta(31)
    PREFERRED_URL_SCHEME = 'http'
    PRESERVE_CONTEXT_ON_EXCEPTION = None
    PROPAGATE_EXCEPTIONS = None
    SECRET_KEY = None
    SEND_FILE_MAX_AGE_DEFAULT = 43200
    SERVER_NAME = None
    SESSION_COOKIE_DOMAIN = None
    SESSION_COOKIE_HTTPONLY = True
    SESSION_COOKIE_NAME = 'session'
    SESSION_COOKIE_PATH = None
    SESSION_COOKIE_SECURE = False
    TESTING = False
    TRAP_BAD_REQUEST_ERRORS = False
    TRAP_HTTP_EXCEPTIONS = False
    USE_X_SENDFILE = False



"""
Data-driven tests for the GA4GH reference implementation. A data
driven test applies a given test method to a data file, and
each applicaton is an independent test case under nose.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import contextlib
import fnmatch
import os
import inspect

import google.protobuf.json_format as json_format

import ga4gh.schemas.pb as pb
import ga4gh.schemas.protocol as protocol


def _wrapTestMethod(method):
    """
    Takes an instance method of a DataDrivenTest subclass, and returns a
    function that can be used in a test generator in nose. This entails
    updating the description attribute so that we can see details of
    the tests being run in nose.
    """
    instance = method.im_self
    cls = instance.__class__

    def testFunction():
        method()
    testFunction.description = "{}.{}.{}:{}".format(
        method.__module__, cls.__name__, method.__name__,
        instance.getLocalId())
    return testFunction


def makeTests(testDataDir, testClass, fnmatchGlob="*"):
    """
    Top-level entry point for data driven tests. For every subdirectory
    in testDataDir, create an instance of testClass and then yield
    each of its testMethods in a format suitable for use with nose
    test generators.
    """
    for localId in os.listdir(testDataDir):
        if (fnmatch.fnmatch(localId, fnmatchGlob) and
                not fnmatch.fnmatch(localId, '*.json') and
                not localId.startswith(".")):  # hidden files start with `.`
            path = os.path.join(testDataDir, localId)
            tester = testClass(localId, path)
            for name, _ in inspect.getmembers(testClass):
                if name.startswith("test"):
                    yield _wrapTestMethod(getattr(tester, name))


class TestCase(object):
    """
    Base class for datadriven test classes.
    Contains assert methods.
    """

    # Protocol Buffers default values are considered "None-like"
    consideredNone = [None, pb.DEFAULT_STRING, pb.DEFAULT_INT]

    def assertEqual(self, a, b):
        """
        Tests if a an b are equal. If not, output an error and raise
        an assertion error.
        """
        if a == b:
            return

        # Protocol Buffers has default string as "", not None
        if a in TestCase.consideredNone and b in TestCase.consideredNone:
            return

        raise AssertionError("{} != {}".format(a, b))

    def assertNotEqual(self, a, b):
        """
        Tests if a and b are not equal.  If they are equal, raise
        an assertion error.
        """
        if a == b:
            raise AssertionError("{} == {}".format(a, b))

    def assertTrue(self, x):
        """
        Tests that x is true.  If it is false, raise an assertion error.
        """
        if not bool(x):
            raise AssertionError("{} is False".format(x))

    def assertFalse(self, x):
        """
        Tests that x is false.  If it is true, raise an assertion error.
        """
        if bool(x):
            raise AssertionError("{} is True".format(x))

    def assertIs(self, a, b):
        """
        Tests that a is b.  If it is not, raise an assertion error.
        """
        if a is not b:
            raise AssertionError("{} is not {}".format(a, b))

    def assertIsNot(self, a, b):
        """
        Tests that a is not b.  If a is b, raise an assertion error.
        """
        if a is b:
            raise AssertionError("{} is {}".format(a, b))

    def assertIsNone(self, x):
        """
        Tests that x is None.  If x is not, raise an assertion error.
        """
        if x not in TestCase.consideredNone:
            raise AssertionError("{} is not None".format(x))

    def assertIsNotNone(self, x):
        """
        Tests that x is not None.  If x is None, raise an assertion error.
        """
        if x in TestCase.consideredNone:
            raise AssertionError("{} is None".format(x))

    def assertIn(self, a, b):
        """
        Tests that a is in b.  If a is not in b, raise an assertion error.
        """
        if a not in b:
            raise AssertionError("{} not in {}".format(a, b))

    def assertNotIn(self, a, b):
        """
        Tests that a is not in b.  If a is in b, raise an assertion error.
        """
        if a in b:
            raise AssertionError("{} in {}".format(a, b))

    def assertIsInstance(self, a, b):
        """
        Tests that a is an instance of b.  If a is not an instance of b,
        raise an assertion error.
        """
        if not isinstance(a, b):
            raise AssertionError("{} is not instance of {}".format(a, b))

    def assertNotIsInstance(self, a, b):
        """
        Tests that a is not an instance of b.  If a is an instance of b,
        raise an assertion error.
        """
        if isinstance(a, b):
            raise AssertionError("{} is instance of {}".format(a, b))

    def assertGreater(self, a, b):
        """
        Tests that a is greater than b.  If not, raise an assertion error.
        """
        if a <= b:
            raise AssertionError("{} is not greater than {}".format(a, b))

    def assertGreaterEqual(self, a, b):
        """
        Tests that a is greater than or equal to b.
        If not, raise an assertion error.
        """
        if a < b:
            raise AssertionError(
                "{} is not greater than or equal to {}".format(a, b))

    def assertLess(self, a, b):
        """
        Tests that a is less than b.  If not, raise an assertion error.
        """
        if a >= b:
            raise AssertionError("{} is not less than {}".format(a, b))

    def assertLessEqual(self, a, b):
        """
        Tests that a is less than or equal to b.
        If not, raise an assertion error.
        """
        if a > b:
            raise AssertionError(
                "{} is not less than or equal to {}".format(a, b))

    def assertRaises(self, exceptionType, func=None, *args, **kwargs):
        """
        Tests that calling func raises an exception of type exceptionType
        """
        if func is None:
            return self.assertRaisesWith(exceptionType)
        exceptionRaised = False
        try:
            func(*args, **kwargs)
        except exceptionType:
            exceptionRaised = True
        if not exceptionRaised:
            raise AssertionError(
                "exception of type {} not raised".format(
                    exceptionType.__name__))

    @contextlib.contextmanager
    def assertRaisesWith(self, exceptionType):
        """
        Tests that a block of code raises an exception of type exceptionType
        """
        exceptionRaised = False
        try:
            yield
        except exceptionType:
            exceptionRaised = True
        if not exceptionRaised:
            raise AssertionError(
                "exception of type {} not raised".format(
                    exceptionType.__name__))

    def assertAlmostEqual(self, a, b, ndigits=7):
        """
        Assert that a and b are equal within ndigits digits
        """
        if round(a-b, ndigits) != 0:
            message = "{} and {} not equal within {} digits".format(
                a, b, ndigits)
            raise AssertionError(message)

    def assertNotAlmostEqual(self, a, b, ndigits=7):
        """
        Assert that a and b are not equal within ndigits digits
        """
        if round(a-b, ndigits) == 0:
            message = "{} and {} equal within {} digits".format(
                a, b, ndigits)
            raise AssertionError(message)


class DataDrivenTest(TestCase):
    """
    Superclass of all data driven tests for GA4GH datamodel objects.
    A data driven test class is instantiated with a set of data files
    that represent a some aggregation of data (for example, a
    ReferenceSet or VariantSet). We allocate a GA4GH datamodel object
    corresponding to this, and then test that these objects have the
    properties that we expect.
    """
    def __init__(self, localId, dataPath):
        self._localId = localId
        self._dataPath = dataPath
        self._gaObject = self.getDataModelInstance(localId, dataPath)

    def getLocalId(self):
        """
        Return the ID of this GA4GH datamodel object we are testing.
        """
        return self._localId

    def getDataModelInstance(self, localId, dataPath):
        """
        Returns the GA4GH datamodel class that this data driven test
        is exercising.
        """
        raise NotImplementedError()

    def getProtocolClass(self):
        """
        Returns the GA4GH protocol class that this data driven test
        is exercising.
        """
        raise NotImplementedError()

    def assertValid(self, protocolClass, jsonDict):
        """
        Asserts that the specified JSON dictionary is a valid instance
        of the specified protocol class.
        """
        try:
            json_format.Parse(jsonDict, protocolClass())
        except json_format.ParseError, e:
            assert False, e.message

    def testProtocolElementValid(self):
        self.assertValid(
            self.getProtocolClass(),
            protocol.toJson(self._gaObject.toProtocolElement()))



"""
A script that takes the compliance dataset (the released version
of which is at https://github.com/ga4gh/compliance/tree/master/test-data)
and turns it into a directory bundle of binary and JSON files suitable
for use by the reference server.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import os
import shutil
import json
import pysam
import generate_gff3_db
import tempfile
import zipfile
import glob

import file_downloader

import ga4gh.common.utils as utils
import glue

glue.ga4ghImportGlue()

# We need to turn off QA because of the import glue
import ga4gh.server  # NOQA
import ga4gh.server.datarepo as datarepo  # NOQA
import ga4gh.server.datamodel.references as references  # NOQA
import ga4gh.server.datamodel.datasets as datasets  # NOQA
import ga4gh.server.datamodel.variants as variants  # NOQA
import ga4gh.server.datamodel.reads as reads  # NOQA
import ga4gh.server.datamodel.ontologies as ontologies  # NOQA
import ga4gh.server.datamodel.sequence_annotations as sequence_annotations  # NOQA
import ga4gh.server.datamodel.continuous as continuous # NOQA
import ga4gh.server.datamodel.bio_metadata as biodata  # NOQA
import ga4gh.server.datamodel.genotype_phenotype_featureset as g2p_featureset  # NOQA
import ga4gh.server.datamodel.genotype_phenotype as g2p_associationset  # NOQA
import ga4gh.server.datamodel.rna_quantification as rna_quantification  # NOQA
import ga4gh.server.repo.rnaseq2ga as rnaseq2ga  # NOQA


class ComplianceDataMunger(object):

    def __init__(self, inputDirectory, outputDirectory, force):
        """
        Converts human readable dataset from compliance repository,
        and translates it into a reference-server readable filesystem
        with binary files.
        :param inputDirectory: location of
            the human readable compliance dataset
        :param outputDirectory: location of
            the file hierarchy suitable for deploying on the reference server
        """
        self.inputDirectory = inputDirectory
        self.outputDirectory = outputDirectory
        self.repoPath = os.path.abspath(
            os.path.join(outputDirectory, "registry.db"))
        self.tempdir = None

        if os.path.exists(self.outputDirectory):
            if force:
                utils.log(
                    "Removing existing output directory at '{}'".format(
                        self.outputDirectory))
                shutil.rmtree(self.outputDirectory)
            else:
                utils.log(
                    "Output directory '{}' already exists".format(
                        self.outputDirectory))
                utils.log(
                    "Please specify an output path that does not exist")
                utils.log("Exiting...")
                exit(1)

        # If no input directory is specified download from GitHub
        if inputDirectory is None:
            utils.log("Downloading test data...")
            self.tempdir = tempfile.mkdtemp()
            assert(os.path.exists(self.tempdir))
            url = "https://github.com/ga4gh/compliance/archive/master.zip"
            filePath = os.path.join(self.tempdir, 'compliance-master.zip')
            downloader = file_downloader.HttpFileDownloader(url, filePath)
            downloader.download()
            utils.log("Extracting test data...")
            with zipfile.ZipFile(filePath, "r") as z:
                z.extractall(self.tempdir)
            self.inputDirectory = os.path.join(
                self.tempdir, 'compliance-master', 'test-data')
        repo = datarepo.SqlDataRepository(self.repoPath)
        self.repo = repo

    def run(self):
        if not os.path.exists(self.outputDirectory):
            os.makedirs(self.outputDirectory)
        self.repo.open("w")
        self.repo.initialise()

        referenceFileName = "ref_brca1.fa"
        inputRef = os.path.join(
            self.inputDirectory, referenceFileName)
        outputRef = os.path.join(
            self.outputDirectory, referenceFileName)
        shutil.copy(inputRef, outputRef)
        fastaFilePath = os.path.join(
            self.outputDirectory,
            referenceFileName + '.gz')
        pysam.tabix_compress(
            outputRef, fastaFilePath)

        with open(
                os.path.join(
                    self.inputDirectory, "ref_brca1.json")) as refMetadataFile:
            refMetadata = json.load(refMetadataFile)
        with open(
                os.path.join(
                    self.inputDirectory,
                    "referenceset_hg37.json")) as refMetadataFile:
            refSetMetadata = json.load(refMetadataFile)

        referenceSet = references.HtslibReferenceSet(
            refSetMetadata['assemblyId'])

        referenceSet.populateFromFile(os.path.abspath(fastaFilePath))
        referenceSet.setAssemblyId(refSetMetadata['assemblyId'])
        referenceSet.setDescription(refSetMetadata['description'])
        if refSetMetadata['species']:
            speciesJson = json.dumps(refSetMetadata['species'])
            referenceSet.setSpeciesFromJson(speciesJson)  # needs a string
        referenceSet.setIsDerived(refSetMetadata['isDerived'])
        referenceSet.setSourceUri(refSetMetadata['sourceUri'])
        referenceSet.setSourceAccessions(refSetMetadata['sourceAccessions'])
        for reference in referenceSet.getReferences():
            if refSetMetadata['species']:
                speciesJsonStr = json.dumps(refMetadata['species'])
                reference.setSpeciesFromJson(speciesJsonStr)
            reference.setSourceAccessions(
                refMetadata['sourceAccessions'])
        self.repo.insertReferenceSet(referenceSet)

        dataset = datasets.Dataset("brca1")
        # Some info is set, it isn't important what
        dataset.setAttributes({"version": ga4gh.server.__version__})
        self.repo.insertDataset(dataset)

        hg00096Individual = biodata.Individual(dataset, "HG00096")
        with open(
                os.path.join(
                    self.inputDirectory,
                    "individual_HG00096.json")) as jsonString:
            hg00096Individual.populateFromJson(jsonString.read())
        self.repo.insertIndividual(hg00096Individual)
        hg00096Biosample = biodata.Biosample(dataset, "HG00096")
        with open(
                os.path.join(
                    self.inputDirectory,
                    "biosample_HG00096.json")) as jsonString:
            hg00096Biosample.populateFromJson(jsonString.read())
        hg00096Biosample.setIndividualId(hg00096Individual.getId())
        self.repo.insertBiosample(hg00096Biosample)
        hg00099Individual = biodata.Individual(dataset, "HG00099")
        with open(
                os.path.join(
                    self.inputDirectory,
                    "individual_HG00099.json")) as jsonString:
            hg00099Individual.populateFromJson(jsonString.read())
        self.repo.insertIndividual(hg00099Individual)
        hg00099Biosample = biodata.Biosample(dataset, "HG00099")
        with open(
                os.path.join(
                    self.inputDirectory,
                    "biosample_HG00099.json")) as jsonString:
            hg00099Biosample.populateFromJson(jsonString.read())
        hg00099Biosample.setIndividualId(hg00099Individual.getId())
        self.repo.insertBiosample(hg00099Biosample)
        hg00101Individual = biodata.Individual(dataset, "HG00101")
        with open(
                os.path.join(
                    self.inputDirectory,
                    "individual_HG00101.json")) as jsonString:
            hg00101Individual.populateFromJson(jsonString.read())
        self.repo.insertIndividual(hg00101Individual)
        hg00101Biosample = biodata.Biosample(dataset, "HG00101")
        with open(
                os.path.join(
                    self.inputDirectory,
                    "biosample_HG00101.json")) as jsonString:
            hg00101Biosample.populateFromJson(jsonString.read())
        hg00101Biosample.setIndividualId(hg00101Individual.getId())
        self.repo.insertBiosample(hg00101Biosample)
        readFiles = [
            "brca1_HG00096.sam",
            "brca1_HG00099.sam",
            "brca1_HG00101.sam"]

        for readFile in readFiles:
            name = readFile.split('_')[1].split('.')[0]
            readSrc = pysam.AlignmentFile(
                os.path.join(self.inputDirectory, readFile), "r")
            readDest = pysam.AlignmentFile(
                os.path.join(
                    self.outputDirectory,
                    name + ".bam"),
                "wb", header=readSrc.header)
            destFilePath = readDest.filename
            for readData in readSrc:
                readDest.write(readData)
            readDest.close()
            readSrc.close()
            pysam.index(destFilePath)
            readGroupSet = reads.HtslibReadGroupSet(dataset, name)
            readGroupSet.populateFromFile(os.path.abspath(
                destFilePath), os.path.abspath(destFilePath + ".bai"))
            readGroupSet.setReferenceSet(referenceSet)
            dataset.addReadGroupSet(readGroupSet)
            biosamples = [hg00096Biosample, hg00099Biosample, hg00101Biosample]
            for readGroup in readGroupSet.getReadGroups():
                for biosample in biosamples:
                    if biosample.getLocalId() == readGroup.getSampleName():
                        readGroup.setBiosampleId(biosample.getId())
            self.repo.insertReadGroupSet(readGroupSet)

        ontologyMapFileName = "so-xp-simple.obo"
        inputOntologyMap = os.path.join(
            self.inputDirectory, ontologyMapFileName)
        outputOntologyMap = os.path.join(
            self.outputDirectory, ontologyMapFileName)
        shutil.copy(inputOntologyMap, outputOntologyMap)

        sequenceOntology = ontologies.Ontology("so-xp-simple")
        sequenceOntology.populateFromFile(os.path.abspath(outputOntologyMap))
        sequenceOntology._id = "so-xp-simple"
        self.repo.insertOntology(sequenceOntology)
        self.repo.addOntology(sequenceOntology)

        vcfFiles = [
            "brca1_1kgPhase3_variants.vcf",
            "brca1_WASH7P_annotation.vcf",
            "brca1_OR4F_annotation.vcf"]
        for vcfFile in vcfFiles:
            self.addVariantSet(
                vcfFile,
                dataset,
                referenceSet,
                sequenceOntology,
                biosamples)

        # Sequence annotations
        seqAnnFile = "brca1_gencodev19.gff3"
        seqAnnSrc = os.path.join(self.inputDirectory, seqAnnFile)
        seqAnnDest = os.path.join(self.outputDirectory, "gencodev19.db")
        dbgen = generate_gff3_db.Gff32Db(seqAnnSrc, seqAnnDest)
        dbgen.run()
        gencode = sequence_annotations.Gff3DbFeatureSet(dataset, "gencodev19")
        gencode.setOntology(sequenceOntology)
        gencode.populateFromFile(os.path.abspath(seqAnnDest))
        gencode.setReferenceSet(referenceSet)

        self.repo.insertFeatureSet(gencode)

        # Continuous data
        continuousFile = ("wgEncodeCaltechRnaSeqNhekR1x75dTh1014Ilna"
                          "MinusSignalRep1.bigWig")
        continuousFileSrc = os.path.join(
                            self.inputDirectory, continuousFile)
        continuousFileDest = os.path.join(
                            self.outputDirectory, continuousFile)
        shutil.copy(continuousFileSrc, continuousFileDest)
        signalData = continuous.FileContinuousSet(dataset, "signalData")
        signalData.populateFromFile(os.path.abspath(continuousFileDest))
        signalData.setReferenceSet(referenceSet)
        self.repo.insertContinuousSet(signalData)

        # add g2p featureSet
        g2pPath = os.path.join(self.inputDirectory, "cgd")
        # copy all files input directory to output path
        outputG2PPath = os.path.join(
            self.outputDirectory, "cgd")
        os.makedirs(outputG2PPath)
        for filename in glob.glob(os.path.join(g2pPath, '*.*')):
            shutil.copy(filename, outputG2PPath)

        featuresetG2P = g2p_featureset.PhenotypeAssociationFeatureSet(
            dataset, os.path.abspath(outputG2PPath))
        featuresetG2P.setOntology(sequenceOntology)
        featuresetG2P.setReferenceSet(referenceSet)
        featuresetG2P.populateFromFile(os.path.abspath(outputG2PPath))
        self.repo.insertFeatureSet(featuresetG2P)

        # add g2p phenotypeAssociationSet
        phenotypeAssociationSet = \
            g2p_associationset.RdfPhenotypeAssociationSet(
                dataset, "cgd", os.path.abspath(outputG2PPath))
        self.repo.insertPhenotypeAssociationSet(phenotypeAssociationSet)

        dataset.addFeatureSet(gencode)

        # RNA Quantification
        rnaDbName = os.path.join(self.outputDirectory, "rnaseq.db")
        store = rnaseq2ga.RnaSqliteStore(rnaDbName)
        store.createTables()
        rnaseq2ga.rnaseq2ga(
            self.inputDirectory + "/rna_brca1.tsv",
            rnaDbName, "rna_brca1.tsv", "rsem",
            featureType="transcript",
            readGroupSetNames="HG00096",
            dataset=dataset,
            featureSetNames="gencodev19",
            biosampleId=hg00096Biosample.getId())
        rnaQuantificationSet = rna_quantification.SqliteRnaQuantificationSet(
            dataset, "rnaseq")
        rnaQuantificationSet.setReferenceSet(referenceSet)
        rnaQuantificationSet.populateFromFile(os.path.abspath(rnaDbName))
        self.repo.insertRnaQuantificationSet(rnaQuantificationSet)

    def addVariantSet(
            self, variantFileName, dataset, referenceSet,
            ontology, biosamples):
        inputVcf = os.path.join(
            self.inputDirectory, variantFileName)
        outputVcf = os.path.join(
            self.outputDirectory, variantFileName)
        shutil.copy(inputVcf, outputVcf)
        pysam.tabix_index(outputVcf, preset="vcf")
        variantSet = variants.HtslibVariantSet(
            dataset, variantFileName.split('_')[1])
        variantSet.setReferenceSet(referenceSet)
        variantSet.populateFromFile(
            [os.path.abspath(outputVcf + ".gz")],
            [os.path.abspath(outputVcf + ".gz.tbi")])
        variantSet.checkConsistency()
        for callSet in variantSet.getCallSets():
            for biosample in biosamples:
                if biosample.getLocalId() == callSet.getLocalId():
                    callSet.setBiosampleId(biosample.getId())
        self.repo.insertVariantSet(variantSet)

        for annotationSet in variantSet.getVariantAnnotationSets():
            annotationSet.setOntology(ontology)
            self.repo.insertVariantAnnotationSet(annotationSet)

    def cleanup(self):
        if self.tempdir is not None:
            shutil.rmtree(self.tempdir)
        utils.log("Done converting compliance data.")
        utils.log("Result in '{}'".format(self.outputDirectory))


@utils.Timed()
def main():
    parser = argparse.ArgumentParser(
        description="Script to generate data bundle from a locally stored "
        "(and possibly locally edited) version of the compliance dataset.")
    parser.add_argument(
        "--outputDirectory", "-o", default="ga4gh-compliance-data",
        help="The directory to output the server-ready data bundle to.")
    parser.add_argument(
        "--inputDirectory", "-i",
        help="Path to local directory containing compliance dataset. "
        "If no directory is provided this script will attempt to "
        "download the compliance test-data from github",
        default=None)
    parser.add_argument('--verbose', '-v', action='count', default=0)
    parser.add_argument('--force', '-f', action='store_true', default=False)
    args = parser.parse_args()
    cdm = ComplianceDataMunger(
        args.inputDirectory, args.outputDirectory, args.force)
    try:
        cdm.run()
    finally:
        cdm.cleanup()


if __name__ == "__main__":
    main()



"""
Unit tests for the oidc code in the frontend.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest
import socket
import mock
import urlparse
import logging
import mimetypes
import shutil

import oic
import oic.oic.message as message

import ga4gh.server.frontend as frontend
import ga4gh.common.utils as utils


RANDSTR = 'abc'
OICCODE = (
    "5GzAsW4Gfz8Km8PTuayFi6j2tA2VnPrHiWybL%2BkryEpr05D0OUKTGTi49Z54yl%2"
    "FIB5KsUzGwSg%2BFpx%2FpBt0aVc3QsvPlsJu6WGHIchJ55UU%3D")

OICRESPONSE = {
    'scope': 'openid profile',
    'state': RANDSTR,
    'code': '5GzAsW4Gfz8Km8PTuayFi6j2tA2VnPrHiWybL+'
            'kryEpr05D0OUKTGTi49Z54yl/IB5KsUzGwSg+'
            'Fpx/pBt0aVc3QsvPlsJu6WGHIchJ55UU='}

# see https://urllib3.readthedocs.org/en/latest/security.html
# #insecurerequestwarning
# for explanation
logging.captureWarnings(True)


def mockAuthorizationRequest(self, request_args, state):
    class mockResult:
        def __init__(self):
            self.url = None

    result = mockResult()
    urlString = 'http://auth.com/auth?nonce={0}&state={1}&redirect_uri={2}'
    result.url = urlString.format(
        request_args['nonce'], state, request_args['redirect_uri'])
    return result


def mockRndstr(size):
    return RANDSTR


ARESP = {'scope': u'openid profile', 'state': RANDSTR,
         'code': OICRESPONSE['code']}


def mockParseResponse(self, response_class, info, sformat):
    response = message.AuthorizationResponse()
    response.from_dict(ARESP)
    return response


ATR = {
    'access_token': '5GzAsW4Gfz8Km8PTuayFi6j2tA2VnPrHiWybL+kryEpr05D0OUKT'
                    'GTi49Z54yl/ImAwqKJ28BWaX6Bvzwj+7uD4udUz'
                    'gFh7bRlFXaQsjz+M=',
    'id_token': {'nonce': RANDSTR,
                 'sub': '-2c68a9e99c6eb6da',
                 'iss': 'https://localhost:8443/',
                 'acr': 'password',
                 'exp': 1435138548,
                 'auth_time': 1435055251,
                 'iat': 1435055748,
                 'aud': ['aKRswYgOAiye']},
    'expires_in': 3600,
    'token_type': 'Bearer',
    'state': RANDSTR,
    'scope': 'openid profile',
    'refresh_token': '5GzAsW4Gfz8Km8PTuayFi6j2tA2VnPrHiWybL+kryEpr05D0OUK'
                     'TGTi49Z54yl/IRkog8/WippmBGCn5UR1'
                     '/W3fxa3wuWQxcjL+4B9BoqHY='}


def mockAccessTokenRequest(self, scope, state, request_args):
    response = message.AccessTokenResponse()
    response.from_dict(ATR)
    return response


def mockAccessTokenRequestError(self, scope, state, request_args):
    return 'random string that is not a message.AccessTokenResponse'


class TestFrontendOidc(unittest.TestCase):
    """
    Tests the OIDC support for the Flask app.
    """

    @classmethod
    def setUpClass(cls):
        config = {
            "SIMULATED_BACKEND_RANDOM_SEED": 1111,
            "SIMULATED_BACKEND_NUM_CALLS": 1,
            "SIMULATED_BACKEND_VARIANT_DENSITY": 1.0,
            "SIMULATED_BACKEND_NUM_VARIANT_SETS": 1,
            "DATA_SOURCE": "simulated://",
            "OIDC_CLIENT_ID": "123",
            "OIDC_CLIENT_SECRET": RANDSTR,
            "OIDC_PROVIDER": "https://accounts.example.com",
            "SECRET_KEY": "secret"
            # "DEBUG" : True
        }
        shutil.rmtree('/tmp/ga4gh', True)
        frontend.reset()
        frontend.configure(
            baseConfig="TestOidcConfig", extraConfig=config, port=8001)
        cls.app = frontend.app.test_client()

    @classmethod
    def tearDownClass(cls):
        cls.app = None
        shutil.rmtree('/tmp/ga4gh', True)

    def sendPostRequest(self, path, request):
        """
        Sends the specified GA request object and returns the response.
        """
        versionedPath = utils.applyVersion(path)
        headers = {
            'Content-type': 'application/json',
            'Origin': self.exampleUrl,
        }
        return self.app.post(
            versionedPath, headers=headers,
            data=request.toJsonString())

    @mock.patch('oic.oauth2.rndstr', mockRndstr)
    @mock.patch('oic.oic.Client.do_authorization_request',
                mockAuthorizationRequest)
    def testAuthorizationRequest(self):
        """
        Test that when the index is referenced, we get a redirect to the
        authorization provider, of the correct form.
        """
        response = self.app.get('/')
        scheme, netloc, path, params, query, fragment = urlparse.urlparse(
            response.location)
        self.assertEqual(netloc, 'auth.com')
        args = urlparse.parse_qs(query, strict_parsing=True)
        self.assertEqual(path, '/auth')
        self.assertEqual(args['redirect_uri'][0],
                         'https://{}:8001/oauth2callback'.format(
                             socket.gethostname()))

    @mock.patch('oic.oauth2.rndstr', lambda x: RANDSTR)
    @mock.patch('oic.oic.Client.parse_response',
                mockParseResponse)
    @mock.patch('oic.oic.Client.do_access_token_request',
                mockAccessTokenRequest)
    def testOidcCallbackSucceeds(self):
        """
        Test that when the authorization provider calls back to us, we can
        handle the reply correctly. This test is the 'succeeds' case.
        """
        url = '/oauth2callback?scope=openid+profile&state={0}&code={1}'.format(
            oic.oauth2.rndstr(0), OICCODE
        )
        with self.app.session_transaction() as sess:
            sess['state'] = oic.oauth2.rndstr(0)
            sess['nonce'] = oic.oauth2.rndstr(0)
        result = self.app.get(url)
        self.assertEqual(result.status_code, 302)
        self.assertEqual(result.location, 'http://{0}:8001/'.format(
            socket.gethostname()))

    def testSessionKeyAllowsIndex(self):
        """
        Test that if we have a valid session in play, we retrieve the index
        page
        """
        with self.app as app:
            with app.session_transaction() as sess:
                sess['key'] = 'xxx'
            app.application.cache.set('xxx', RANDSTR)
            result = app.get('/')
            self.assertEqual(result.status_code, 200)
            self.assertEqual("text/html", result.mimetype)
            self.assertGreater(len(result.data), 0)

    def testKeyParamAllowsIndex(self):
        """
        Test that if we have a valid key in play, we retrieve the index
        page
        """
        with self.app as app:
            app.application.cache.set('xxx', RANDSTR)
            result = app.get('/?key=xxx')
            self.assertEqual(result.status_code, 200)
            self.assertEqual("text/html", result.mimetype)
            self.assertGreater(len(result.data), 0)

    @mock.patch('oic.oauth2.rndstr', mockRndstr)
    @mock.patch('oic.oic.Client.do_authorization_request',
                mockAuthorizationRequest)
    def testInvalidSessionKeyRedirects(self):
        """
        Test that if we have an invalid session in play, we redirected to OP
        """
        with self.app as app:
            with app.session_transaction() as sess:
                sess['key'] = 'xxx'
            result = app.get('/')
            scheme, netloc, path, params, query, fragment = urlparse.urlparse(
                result.location)
            self.assertEqual(netloc, 'auth.com')
            args = urlparse.parse_qs(query, strict_parsing=True)
            self.assertEqual(path, '/auth')
            self.assertEqual(args['redirect_uri'][0],
                             'https://{}:8001/oauth2callback'.format(
                                 socket.gethostname()))

    @mock.patch('oic.oauth2.rndstr', mockRndstr)
    @mock.patch('oic.oic.Client.do_authorization_request',
                mockAuthorizationRequest)
    def testInvalidKeyParamIsUnauthorized(self):
        """
        Test that if we have an invalid key in play, we get an exception
        """
        result = self.app.get('/?key=xxx')
        self.assertEqual(result.status_code, 403)

    @mock.patch('oic.oauth2.rndstr', lambda x: RANDSTR)
    @mock.patch('oic.oic.Client.parse_response',
                mockParseResponse)
    @mock.patch('oic.oic.Client.do_access_token_request',
                mockAccessTokenRequest)
    def testOidcCallbackBadNonce(self):
        """
        Test that when the authorization provider calls back to us, we can
        handle the reply correctly.
        In this case, the nonce returned does not match that in the session.
        """
        url = '/oauth2callback?scope=openid+profile&state={0}&code={1}'.format(
            oic.oauth2.rndstr(0), OICCODE
        )
        with self.app as app:
            with app.session_transaction() as sess:
                sess['state'] = oic.oauth2.rndstr(0)
                sess['nonce'] = 'other'
            result = app.get(url)
            self.assertEqual(result.status_code, 403)

    @mock.patch('oic.oauth2.rndstr', lambda x: RANDSTR)
    @mock.patch('oic.oic.Client.parse_response',
                mockParseResponse)
    @mock.patch('oic.oic.Client.do_access_token_request',
                mockAccessTokenRequest)
    def testOidcCallbackBadState(self):
        """
        Test that when the authorization provider calls back to us, we can
        handle the reply correctly.
        In this case, the state returned does not match that in the session.
        """
        url = '/oauth2callback?scope=openid+profile&state={0}&code={1}'.format(
            oic.oauth2.rndstr(0), OICCODE
        )
        with self.app as app:
            with app.session_transaction() as sess:
                sess['state'] = 'other'
                sess['nonce'] = oic.oauth2.rndstr(0)
            result = app.get(url)
            self.assertEqual(result.status_code, 403)

    @mock.patch('oic.oauth2.rndstr', lambda x: RANDSTR)
    @mock.patch('oic.oic.Client.parse_response',
                mockParseResponse)
    @mock.patch('oic.oic.Client.do_access_token_request',
                mockAccessTokenRequestError)
    def testOidcCallbackBadTokenResponse(self):
        """
        Test that when the authorization provider calls back to us, we can
        handle the reply correctly.
        In this case, the access token request returns an invalid response.
        """
        url = '/oauth2callback?scope=openid+profile&state={0}&code={1}'.format(
            oic.oauth2.rndstr(0), OICCODE
        )
        with self.app as app:
            with app.session_transaction() as sess:
                sess['state'] = oic.oauth2.rndstr(0)
                sess['nonce'] = oic.oauth2.rndstr(0)
            result = app.get(url)
            self.assertEqual(result.status_code, 403)

    def testMimetype(self):
        """
        The  `oidc-provider` relies on mimetypes.guess_type(full_path)
        `mimetypes` in turn relies on the host OS configuration,
        typically ["/etc/mime.types","/etc/httpd/conf/mime.types",...]
        If this test fails, see the results of `mimetypes.knownfiles`
        on your host and ensure configuration set up
        see https://github.com/ga4gh/ga4gh-server/issues/501
        """
        content_type, encoding = mimetypes.guess_type("foo.json")
        self.assertEqual(content_type, "application/json")



"""
Data-driven tests for references.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import hashlib
import os
import unittest

# TODO it may be a bit circular to use pysam as our interface for
# accessing reference information, since this is the method we use
# in the main code. We should examine the possibility of using
# a different library here --- perhaps BioPython?
import pysam

import ga4gh.server.datamodel.references as references
import ga4gh.server.exceptions as exceptions
import tests.datadriven as datadriven
import tests.paths as paths

import ga4gh.schemas.protocol as protocol


def testReferenceSets():
    testDataDir = os.path.join(paths.testDataDir, "referenceSets")
    pattern = "*.fa.gz"
    for test in datadriven.makeTests(testDataDir, ReferenceSetTest, pattern):
        yield test


class ReferenceSetTest(datadriven.DataDrivenTest):
    """
    Data drive test class for reference sets. Builds an alternative model of
    a reference set, and verifies that it is consistent with the model
    built by the references.ReferenceSet object.
    """
    def __init__(self, referenceSetId, fastaFile):
        super(ReferenceSetTest, self).__init__(referenceSetId, fastaFile)
        self._fastaFile = pysam.FastaFile(fastaFile)

    def getDataModelInstance(self, localId, dataPath):
        referenceSet = references.HtslibReferenceSet(localId)
        referenceSet.populateFromFile(dataPath)
        return referenceSet

    def getProtocolClass(self):
        return protocol.ReferenceSet

    def testValidateObjects(self):
        # test that validation works on reference sets and references
        referenceSet = self._gaObject
        referenceSetPe = referenceSet.toProtocolElement()
        self.assertValid(
            protocol.ReferenceSet, protocol.toJson(referenceSetPe))
        for gaReference in referenceSet.getReferences():
            reference = protocol.toJson(gaReference.toProtocolElement())
            self.assertValid(protocol.Reference, reference)

    def testGetBases(self):
        # test searching with no arguments succeeds
        referenceSet = self._gaObject
        for gaReference in referenceSet.getReferences():
            self.assertReferencesEqual(gaReference)

    def testGetBasesStart(self):
        # test searching with start only succeeds
        self.doRangeTest(5, None)

    def testGetBasesEnd(self):
        # test searching with end only succeeds
        self.doRangeTest(None, 5)

    def testGetBasesRanges(self):
        # test searching with start and end succeeds
        self.doRangeTest(2, 5)

    @unittest.skip("We assume that the 0 is unset, as protobuf3 has 0 == None")
    def testGetBasesEmpty(self):
        self.doRangeTest(0, 0)

    def testOutOfBounds(self):
        referenceSet = self._gaObject
        for reference in referenceSet.getReferences():
            length = reference.getLength()
            badRanges = [
                (-1, 1), (0, length + 1), (0, 2**34), (-2**32, 1),
                (1, 0), (length, length - 1),
            ]
            for start, end in badRanges:
                self.assertRaises(
                    exceptions.ReferenceRangeErrorException,
                    reference.getBases, start, end)

    def testMd5checksums(self):
        referenceSet = self._gaObject
        referenceMd5s = []
        for gaReference in referenceSet.getReferences():
            bases = self._fastaFile.fetch(gaReference.getLocalId())
            basesChecksum = hashlib.md5(bases).hexdigest()
            self.assertEqual(basesChecksum, gaReference.getMd5Checksum())
            referenceMd5s.append(gaReference.getMd5Checksum())
        referenceMd5s.sort()
        checksumsString = ''.join(referenceMd5s)
        md5checksum = hashlib.md5(checksumsString).hexdigest()
        referenceSetMd5 = referenceSet.getMd5Checksum()
        self.assertEqual(md5checksum, referenceSetMd5)

    def doRangeTest(self, start=None, end=None):
        referenceSet = self._gaObject
        for gaReference in referenceSet.getReferences():
            self.assertReferencesEqual(gaReference, start, end)

    def assertReferencesEqual(
            self, gaReference, start=None, end=None):
        theStart = 0 if start is None else start
        bases = self._fastaFile.fetch(gaReference.getLocalId())
        theEnd = len(bases) if end is None else end
        gaBases = gaReference.getBases(theStart, theEnd)
        pysamBases = bases[theStart:theEnd]
        self.assertEqual(gaBases, pysamBases)



"""
Support for Ontologies.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections
import os.path

import ga4gh.server.exceptions as exceptions
import ga4gh.server.datamodel.obo_parser as obo_parser

import ga4gh.schemas.protocol as protocol


SEQUENCE_ONTOLOGY_PREFIX = "SO"


class OboReader(obo_parser.OBOReader):
    """
    We extend the OBOReader class to allow us throw a custom exception
    so that it will be handled correctly. The default implementation
    throws an Exception instance on error, which cannot be caught
    without masking pretty much any kind of error.
    """
    def _die(self, message, line):
        raise exceptions.OntologyFileFormatException(
            self.obo_file, "Error at line {}: {}".format(line, message))


class Ontology(object):
    """
    A bidectional map between ontology names and IDs (e.g. in Sequence
    Ontology we would have "SO:0001583 <-> missense_variant") derived
    from an OBO file.
    """
    def __init__(self, name):
        self._id = None
        self._name = name
        self._sourceVersion = None
        self._ontologyPrefix = None
        self._dataUrl = None
        # There can be duplicate names, so we need to store a list of IDs.
        self._nameIdMap = collections.defaultdict(list)

    def _readFile(self):
        if not os.path.exists(self._dataUrl):
            raise exceptions.FileOpenFailedException(self._dataUrl)
        reader = OboReader(obo_file=self._dataUrl)
        ids = set()
        for record in reader:
            if record.id in ids:
                raise exceptions.OntologyFileFormatException(
                    self._dataUrl, "Duplicate ID {}".format(record.id))
            ids.add(record.id)
            self._nameIdMap[record.name].append(record.id)
        self._sourceVersion = reader.format_version
        if len(ids) == 0:
            raise exceptions.OntologyFileFormatException(
                self._dataUrl, "No valid records found.")
        # To get prefix, pull out an ID and parse it.
        self._ontologyPrefix = record.id.split(":")[0]
        self._sourceVersion = reader.data_version

    def populateFromFile(self, dataUrl):
        """
        Populates this ontology map from the specified dataUrl.
        This reads the ontology term name and ID pairs from the
        specified file.
        """
        self._dataUrl = dataUrl
        self._readFile()

    def populateFromRow(self, ontologyRecord):
        """
        Populates this Ontology using values in the specified DB row.
        """
        self._id = ontologyRecord.id
        self._dataUrl = ontologyRecord.dataurl
        self._readFile()
        # TODO sanity check the stored values against what we have just read.

    def getId(self):
        """
        Returns the ID of this Ontology. This is an internal
        identifier.
        """
        return self._id

    def getOntologyPrefix(self):
        """
        Returns the ontology prefix string, i.e. "SO" for a sequence
        ontology and "GO" for gene a ontology.
        """
        return self._ontologyPrefix

    def getDataUrl(self):
        return self._dataUrl

    def getName(self):
        """
        Returns the name of this ontology.
        """
        return self._name

    def getTermIds(self, termName):
        """
        Returns the list of ontology IDs scorresponding to the specified term
        name. If the term name is not found, return the empty list.
        """
        return self._nameIdMap[termName]

    def getGaTermByName(self, name):
        """
        Returns a GA4GH OntologyTerm object by name.

        :param name: name of the ontology term, ex. "gene".
        :return: GA4GH OntologyTerm object.
        """
        # TODO what is the correct value when we have no mapping??
        termIds = self.getTermIds(name)
        if len(termIds) == 0:
            termId = ""
            # TODO add logging for missed term translation.
        else:
            # TODO what is the correct behaviour here when we have multiple
            # IDs matching a given name?
            termId = termIds[0]
        term = protocol.OntologyTerm()
        term.term = name
        term.term_id = termId
        return term



"""
Simple script to decode exception error codes. This translates
the error code received by clients into an exception class.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse

import ga4gh.server.exceptions as exceptions


def parseArgs():
    parser = argparse.ArgumentParser(
        description=(
            "Converts an error code received by a clients to the  "
            "corresponding exception class."))
    parser.add_argument(
        "errorCode", type=int,
        help="The errorCode value in a GAException object.")
    args = parser.parse_args()
    return args


def main():
    args = parseArgs()
    exceptionClass = exceptions.getExceptionClass(args.errorCode)
    print(args.errorCode, exceptionClass, sep="\t")


if __name__ == '__main__':
    main()



from email.mime.text import MIMEText
import hashlib
import json
import smtplib
import time

from oic.utils.http_util import Response

from provider.authn import AuthnModule, make_cls_from_name
from provider.authn.user_pass import UserPass


class MailTwoFactor(AuthnModule):
    url_endpoint = "/two_factor/verify"

    def __init__(self, user_db, passwd_db, smtp_server, outgoing_sender,
                 template_env, code_ttl=2, template="mail_two_factor.jinja2",
                 **kwargs):
        """

        :param user_db:
        :param smtp_server:
        :param outgoing_sender:
        :param code_ttl: how long the code is valid (in minutes)
        :param kwargs:
        :return:
        """
        super(MailTwoFactor, self).__init__(None)
        self.template_env = template_env
        self.template = template

        self.first_factor = UserPass(passwd_db, template_env)
        self.first_factor.url_endpoint = self.url_endpoint

        cls = make_cls_from_name(user_db["class"])
        self.user_db = cls(**user_db["kwargs"])
        self.smtp_server = smtp_server
        self.outgoing_sender = outgoing_sender
        self.code_ttl = code_ttl * 60 * 1000  # ttl in ms

        self.codes = {}

    def __call__(self, *args, **kwargs):
        # delegate to first factor module
        return self.first_factor(*args, **kwargs)

    def verify(self, *args, **kwargs):
        # second auth with code
        if "code" in kwargs:
            code = kwargs["code"]

            try:
                entry = self.codes[code]
            except KeyError:
                return self.FAILED_AUTHN

            username = entry["username"]
            # Verify code is not too old
            now = time.time()
            if now - entry["time"] > self.code_ttl:
                return self.FAILED_AUTHN

            del self.codes[code]  # Removed used code
            return username, True

        else:
            result = self.first_factor.verify(*args, **kwargs)
            if result == self.FAILED_AUTHN:
                return self.FAILED_AUTHN

            username, _ = result
            try:
                receiver = self.user_db[username]["email"]
            except KeyError:
                # Missing user or no mail address
                self.FAILED_AUTHN

            # Generate code and send it
            code = hashlib.md5(str(time.time())).hexdigest()
            self.codes[code] = {"username": username, "time": time.time()}
            self._send_mail(code, receiver)

            template = self.template_env.get_template(self.template)
            response = Response(template.render(mail=receiver,
                                                action=self.url_endpoint,
                                                state=json.dumps(
                                                    kwargs["state"])))
            return response, False

    def _send_mail(self, code, receiver):
        msg = MIMEText("Code: {}".format(code))
        msg["Subject"] = "Authentication code"
        msg["From"] = self.outgoing_sender
        msg["To"] = receiver

        s = smtplib.SMTP(self.smtp_server)
        s.sendmail(self.outgoing_sender, [receiver], msg.as_string())
        s.quit()



"""
Unit tests for faulty data sets.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import unittest

import ga4gh.server.datamodel.datasets as datasets
import ga4gh.server.datamodel.variants as variants
import ga4gh.server.exceptions as exceptions

import ga4gh.schemas.protocol as protocol


class FaultyVariantDataTest(unittest.TestCase):
    """
    Superclass of faulty variant data tests.
    """
    def setUp(self):
        self.testDataDir = "tests/faultydata/variants"
        self.dataset = datasets.Dataset('dataset1')

    def getFullPath(self, localId):
        return os.path.join(self.testDataDir, localId)


class TestVariantSetNoIndexedVcf(FaultyVariantDataTest):
    localIds = ["no_indexed_vcf"]

    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            variantSet = variants.HtslibVariantSet(self.dataset, localId)
            with self.assertRaises(exceptions.NotIndexedException):
                variantSet.populateFromDirectory(path)


class TestInconsistentMetaData(FaultyVariantDataTest):
    localIds = ["inconsist_meta"]

    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            variantSet = variants.HtslibVariantSet(self.dataset, localId)
            variantSet.populateFromDirectory(path)
            with self.assertRaises(exceptions.InconsistentMetaDataException):
                variantSet.checkConsistency()


class TestInconsistentCallSetId(FaultyVariantDataTest):
    localIds = ["inconsist_sampleid", "inconsist_sampleid2"]

    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            variantSet = variants.HtslibVariantSet(self.dataset, localId)
            variantSet.populateFromDirectory(path)
            with self.assertRaises(exceptions.InconsistentCallSetIdException):
                variantSet.checkConsistency()


class TestOverlappingVcfVariants(FaultyVariantDataTest):
    localIds = ["overlapping_vcf"]

    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            variantSet = variants.HtslibVariantSet(self.dataset, localId)
            with self.assertRaises(exceptions.OverlappingVcfException):
                variantSet.populateFromDirectory(path)


class TestDuplicateCallSetId(FaultyVariantDataTest):
    """
    THIS SECTION IS CURRENTLY NOT WORKING
    It returns the following error:

    [E::bcf_hdr_add_sample] Duplicated sample name 'S1'
    Aborted (core dumped)

    which is coming from:
    htslib/vcf.c function bcf_hdr_add_sample

    UNABLE TO CAPTURE EXCEPTION
    """
    localIds = ["duplicated_sampleid"]

    @unittest.skipIf(protocol.version.startswith("0.6"), "")
    def testInstantiation(self):
        for localId in self.localIds:
            path = self.getFullPath(localId)
            variantSet = variants.HtslibVariantSet(self.dataset, localId)
            with self.assertRaises(exceptions.DuplicateCallSetIdException):
                variantSet.populateFromDirectory(path)


class FaultyReferenceDataTest(unittest.TestCase):
    """
    Superclass of faulty reference data tests
    """
    def getFullPath(self, localId):
        testDataDir = "tests/faultydata/references"
        return os.path.join(testDataDir, localId)



"""
Dataset objects
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import ga4gh.server.datamodel as datamodel
import ga4gh.server.datamodel.reads as reads
import ga4gh.server.datamodel.sequence_annotations as sequence_annotations
import ga4gh.server.datamodel.continuous as continuous
import ga4gh.server.datamodel.variants as variants
import ga4gh.server.exceptions as exceptions
import ga4gh.server.datamodel.bio_metadata as biodata
import ga4gh.server.datamodel.genotype_phenotype as g2p
import ga4gh.server.datamodel.rna_quantification as rnaQuantification

import ga4gh.schemas.pb as pb
import ga4gh.schemas.protocol as protocol


class Dataset(datamodel.DatamodelObject):
    """
    The base class of datasets containing variants and reads
    """
    compoundIdClass = datamodel.DatasetCompoundId

    def __init__(self, localId):
        super(Dataset, self).__init__(None, localId)
        self._description = None
        self._variantSetIds = []
        self._variantSetIdMap = {}
        self._variantSetNameMap = {}
        self._featureSetIds = []
        self._featureSetIdMap = {}
        self._featureSetNameMap = {}
        self._continuousSetIds = []
        self._continuousSetIdMap = {}
        self._continuousSetNameMap = {}
        self._readGroupSetIds = []
        self._readGroupSetIdMap = {}
        self._readGroupSetNameMap = {}
        self._biosampleIds = []
        self._biosampleIdMap = {}
        self._biosampleNameMap = {}
        self._individualIds = []
        self._individualIdMap = {}
        self._individualNameMap = {}
        self._phenotypeAssociationSetIdMap = {}
        self._phenotypeAssociationSetNameMap = {}
        self._phenotypeAssociationSetIds = []
        self._rnaQuantificationSetIds = []
        self._rnaQuantificationSetIdMap = {}
        self._rnaQuantificationSetNameMap = {}

    def populateFromRow(self, dataset):
        """
        Populates the instance variables of this Dataset from the
        specified database row.
        """
        self._description = dataset.description
        self.setAttributesJson(dataset.attributes)

    def setDescription(self, description):
        """
        Sets the description for this dataset to the specified value.
        """
        self._description = description

    def addVariantSet(self, variantSet):
        """
        Adds the specified variantSet to this dataset.
        """
        id_ = variantSet.getId()
        self._variantSetIdMap[id_] = variantSet
        self._variantSetNameMap[variantSet.getLocalId()] = variantSet
        self._variantSetIds.append(id_)

    def addBiosample(self, biosample):
        """
        Adds the specified biosample to this dataset.
        """
        id_ = biosample.getId()
        self._biosampleIdMap[id_] = biosample
        self._biosampleIds.append(id_)
        self._biosampleNameMap[biosample.getName()] = biosample

    def addIndividual(self, individual):
        """
        Adds the specified individual to this dataset.
        """
        id_ = individual.getId()
        self._individualIdMap[id_] = individual
        self._individualIds.append(id_)
        self._individualNameMap[individual.getName()] = individual

    def addFeatureSet(self, featureSet):
        """
        Adds the specified featureSet to this dataset.
        """
        id_ = featureSet.getId()
        self._featureSetIdMap[id_] = featureSet
        self._featureSetIds.append(id_)
        name = featureSet.getLocalId()
        self._featureSetNameMap[name] = featureSet

    def addContinuousSet(self, continuousSet):
        """
        Adds the specified continuousSet to this dataset.
        """
        id_ = continuousSet.getId()
        self._continuousSetIdMap[id_] = continuousSet
        self._continuousSetIds.append(id_)
        name = continuousSet.getLocalId()
        self._continuousSetNameMap[name] = continuousSet

    def addReadGroupSet(self, readGroupSet):
        """
        Adds the specified readGroupSet to this dataset.
        """
        id_ = readGroupSet.getId()
        self._readGroupSetIdMap[id_] = readGroupSet
        self._readGroupSetNameMap[readGroupSet.getLocalId()] = readGroupSet
        self._readGroupSetIds.append(id_)

    def addRnaQuantificationSet(self, rnaQuantSet):
        """
        Adds the specified rnaQuantification set to this dataset.
        """
        id_ = rnaQuantSet.getId()
        self._rnaQuantificationSetIdMap[id_] = rnaQuantSet
        self._rnaQuantificationSetIds.append(id_)
        name = rnaQuantSet.getLocalId()
        self._rnaQuantificationSetNameMap[name] = rnaQuantSet

    def toProtocolElement(self):
        dataset = protocol.Dataset()
        dataset.id = self.getId()
        dataset.name = pb.string(self.getLocalId())
        dataset.description = pb.string(self.getDescription())
        self.serializeAttributes(dataset)
        return dataset

    def getVariantSets(self):
        """
        Returns the list of VariantSets in this dataset
        """
        return [self._variantSetIdMap[id_] for id_ in self._variantSetIds]

    def getNumVariantSets(self):
        """
        Returns the number of variant sets in this dataset.
        """
        return len(self._variantSetIds)

    def getVariantSet(self, id_):
        """
        Returns the VariantSet with the specified name, or raises a
        VariantSetNotFoundException otherwise.
        """
        if id_ not in self._variantSetIdMap:
            raise exceptions.VariantSetNotFoundException(id_)
        return self._variantSetIdMap[id_]

    def getVariantSetByIndex(self, index):
        """
        Returns the variant set at the specified index in this dataset.
        """
        return self._variantSetIdMap[self._variantSetIds[index]]

    def getVariantSetByName(self, name):
        """
        Returns a VariantSet with the specified name, or raises a
        VariantSetNameNotFoundException if it does not exist.
        """
        if name not in self._variantSetNameMap:
            raise exceptions.VariantSetNameNotFoundException(name)
        return self._variantSetNameMap[name]

    def addPhenotypeAssociationSet(self, phenotypeAssociationSet):
        """
        Adds the specified g2p association set to this backend.
        """
        id_ = phenotypeAssociationSet.getId()
        self._phenotypeAssociationSetIdMap[id_] = phenotypeAssociationSet
        self._phenotypeAssociationSetNameMap[
            phenotypeAssociationSet.getLocalId()] = phenotypeAssociationSet
        self._phenotypeAssociationSetIds.append(id_)

    def getPhenotypeAssociationSets(self):
        return [self._phenotypeAssociationSetIdMap[id_]
                for id_ in self._phenotypeAssociationSetIdMap]

    def getPhenotypeAssociationSet(self, id_):
        return self._phenotypeAssociationSetIdMap[id_]

    def getPhenotypeAssociationSetByName(self, name):
        if name not in self._phenotypeAssociationSetNameMap:
            raise exceptions.PhenotypeAssociationSetNotFoundException(name)
        return self._phenotypeAssociationSetNameMap[name]

    def getPhenotypeAssociationSetByIndex(self, index):
        return self._phenotypeAssociationSetIdMap[
            self._phenotypeAssociationSetIds[index]]

    def getNumPhenotypeAssociationSets(self):
        """
        Returns the number of reference sets in this data repository.
        """
        return len(self._phenotypeAssociationSetIds)

    def getFeatureSets(self):
        """
        Returns the list of FeatureSets in this dataset
        """
        return [self._featureSetIdMap[id_] for id_ in self._featureSetIds]

    def getNumFeatureSets(self):
        """
        Returns the number of feature sets in this dataset.
        """
        return len(self._featureSetIds)

    def getFeatureSet(self, id_):
        """
        Returns the FeatureSet with the specified id, or raises a
        FeatureSetNotFoundException otherwise.
        """
        if id_ not in self._featureSetIdMap:
            raise exceptions.FeatureSetNotFoundException(id_)
        return self._featureSetIdMap[id_]

    def getFeatureSetByName(self, name):
        """
        Returns the FeatureSet with the specified name, or raises
        an exception otherwise.
        """
        if name not in self._featureSetNameMap:
            raise exceptions.FeatureSetNameNotFoundException(name)
        return self._featureSetNameMap[name]

    def getFeatureSetByIndex(self, index):
        """
        Returns the feature set at the specified index in this dataset.
        """
        return self._featureSetIdMap[self._featureSetIds[index]]

    def getContinuousSets(self):
        """
        Returns the list of ContinuousSets in this dataset
        """
        return [self._continuousSetIdMap[id_]
                for id_ in self._continuousSetIds]

    def getNumContinuousSets(self):
        """
        Returns the number of continuous sets in this dataset.
        """
        return len(self._continuousSetIds)

    def getContinuousSet(self, id_):
        """
        Returns the ContinuousSet with the specified id, or raises a
        ContinuousSetNotFoundException otherwise.
        """
        if id_ not in self._continuousSetIdMap:
            raise exceptions.ContinuousSetNotFoundException(id_)
        return self._continuousSetIdMap[id_]

    def getContinuousSetByName(self, name):
        """
        Returns the ContinuousSet with the specified name, or raises
        an exception otherwise.
        """
        if name not in self._continuousSetNameMap:
            raise exceptions.ContinuousSetNameNotFoundException(name)
        return self._continuousSetNameMap[name]

    def getContinuousSetByIndex(self, index):
        """
        Returns the continuous set at the specified index in this dataset.
        """
        return self._continuousSetIdMap[self._continuousSetIds[index]]

    def getBiosamples(self):
        """
        Returns the list of biosamples in this dataset
        """
        return [self._biosampleIdMap[id_] for id_ in self._biosampleIds]

    def getBiosampleByName(self, name):
        """
        Returns a Biosample with the specified name, or raises a
        BiosampleNameNotFoundException if it does not exist.
        """
        if name not in self._biosampleNameMap:
            raise exceptions.BiosampleNameNotFoundException(name)
        return self._biosampleNameMap[name]

    def getBiosample(self, id_):
        """
        Returns the Biosample with the specified id, or raises
        a BiosampleNotFoundException otherwise.
        """
        if id_ not in self._biosampleIdMap:
            raise exceptions.BiosampleNotFoundException(id_)
        return self._biosampleIdMap[id_]

    def getIndividuals(self):
        """
        Returns the list of individuals in this dataset
        """
        return [self._individualIdMap[id_] for id_ in self._individualIds]

    def getIndividualByName(self, name):
        """
        Returns an individual with the specified name, or raises a
        IndividualNameNotFoundException if it does not exist.
        """
        if name not in self._individualNameMap:
            raise exceptions.IndividualNameNotFoundException(name)
        return self._individualNameMap[name]

    def getIndividual(self, id_):
        """
        Returns the Individual with the specified id, or raises
        a IndividualNotFoundException otherwise.
        """
        if id_ not in self._individualIdMap:
            raise exceptions.IndividualNotFoundException(id_)
        return self._individualIdMap[id_]

    def getNumReadGroupSets(self):
        """
        Returns the number of readgroup sets in this dataset.
        """
        return len(self._readGroupSetIds)

    def getReadGroupSets(self):
        """
        Returns the list of ReadGroupSets in this dataset
        """
        return [self._readGroupSetIdMap[id_] for id_ in self._readGroupSetIds]

    def getReadGroupSetByName(self, name):
        """
        Returns a ReadGroupSet with the specified name, or raises a
        ReadGroupSetNameNotFoundException if it does not exist.
        """
        if name not in self._readGroupSetNameMap:
            raise exceptions.ReadGroupSetNameNotFoundException(name)
        return self._readGroupSetNameMap[name]

    def getReadGroupSetByIndex(self, index):
        """
        Returns the readgroup set at the specified index in this dataset.
        """
        return self._readGroupSetIdMap[self._readGroupSetIds[index]]

    def getReadGroupSet(self, id_):
        """
        Returns the ReadGroupSet with the specified name, or raises
        a ReadGroupSetNotFoundException otherwise.
        """
        if id_ not in self._readGroupSetIdMap:
            raise exceptions.ReadGroupNotFoundException(id_)
        return self._readGroupSetIdMap[id_]

    def getInfo(self):
        """
        Returns the info of this dataset.
        """
        return self._info

    def getDescription(self):
        """
        Returns the free text description of this dataset.
        """
        return self._description

    def getNumRnaQuantificationSets(self):
        """
        Returns the number of rna quantification sets in this dataset.
        """
        return len(self._rnaQuantificationSetIds)

    def getRnaQuantificationSets(self):
        """
        Returns the list of RnaQuantification sets in this dataset
        """
        return [self._rnaQuantificationSetIdMap[id_] for
                id_ in self._rnaQuantificationSetIds]

    def getRnaQuantificationSetByIndex(self, index):
        """
        Returns the rna quantification set at the specified index in this
        dataset.
        """
        return self._rnaQuantificationSetIdMap[
            self._rnaQuantificationSetIds[index]]

    def getRnaQuantificationSetByName(self, name):
        """
        Returns the RnaQuantification set with the specified name, or raises
        an exception otherwise.
        """
        if name not in self._rnaQuantificationSetNameMap:
            raise exceptions.RnaQuantificationSetNameNotFoundException(name)
        return self._rnaQuantificationSetNameMap[name]

    def getRnaQuantificationSet(self, id_):
        """
        Returns the RnaQuantification set with the specified name, or raises
        a RnaQuantificationSetNotFoundException otherwise.
        """
        if id_ not in self._rnaQuantificationSetIdMap:
            raise exceptions.RnaQuantificationSetNotFoundException(id_)
        return self._rnaQuantificationSetIdMap[id_]


class SimulatedDataset(Dataset):
    """
    A simulated dataset
    """
    def __init__(
            self, localId, referenceSet, randomSeed=0,
            numVariantSets=1, numCalls=1, variantDensity=0.5,
            numReadGroupSets=1, numReadGroupsPerReadGroupSet=1,
            numAlignments=1, numFeatureSets=1, numContinuousSets=1,
            numPhenotypeAssociationSets=1,
            numPhenotypeAssociations=2, numRnaQuantSets=2,
            numExpressionLevels=2):
        super(SimulatedDataset, self).__init__(localId)
        self._description = "Simulated dataset {}".format(localId)

        for i in range(numPhenotypeAssociationSets):
            localId = "simPas{}".format(i)
            seed = randomSeed + i
            phenotypeAssociationSet = g2p.SimulatedPhenotypeAssociationSet(
                self, localId, seed, numPhenotypeAssociations)
            self.addPhenotypeAssociationSet(phenotypeAssociationSet)

        # TODO create a simulated Ontology
        # Variants
        for i in range(numVariantSets):
            localId = "simVs{}".format(i)
            seed = randomSeed + i
            variantSet = variants.SimulatedVariantSet(
                self, referenceSet, localId, seed, numCalls, variantDensity)
            callSets = variantSet.getCallSets()
            # Add biosamples
            for callSet in callSets:
                biosample = biodata.Biosample(
                    self, callSet.getLocalId())
                biosample2 = biodata.Biosample(
                    self, callSet.getLocalId() + "2")
                individual = biodata.Individual(
                    self, callSet.getLocalId())
                biosample.setIndividualId(individual.getId())
                biosample2.setIndividualId(individual.getId())
                self.addIndividual(individual)
                self.addBiosample(biosample)
                self.addBiosample(biosample2)
            self.addVariantSet(variantSet)
            variantAnnotationSet = variants.SimulatedVariantAnnotationSet(
                variantSet, "simVas{}".format(i), seed)
            variantSet.addVariantAnnotationSet(variantAnnotationSet)
        # Reads
        for i in range(numReadGroupSets):
            localId = 'simRgs{}'.format(i)
            seed = randomSeed + i
            readGroupSet = reads.SimulatedReadGroupSet(
                self, localId, referenceSet, seed,
                numReadGroupsPerReadGroupSet, numAlignments)
            for rg in readGroupSet.getReadGroups():
                biosample = biodata.Biosample(
                    self, rg.getLocalId())
                individual = biodata.Individual(
                    self, rg.getLocalId())
                biosample.setIndividualId(individual.getId())
                rg.setBiosampleId(biosample.getId())
                self.addIndividual(individual)
                self.addBiosample(biosample)
            self.addReadGroupSet(readGroupSet)
        # Features
        for i in range(numFeatureSets):
            localId = "simFs{}".format(i)
            seed = randomSeed + i
            featureSet = sequence_annotations.SimulatedFeatureSet(
                self, localId, seed)
            featureSet.setReferenceSet(referenceSet)
            self.addFeatureSet(featureSet)
        # Continuous
        for i in range(numContinuousSets):
            localId = "simConts{}".format(i)
            seed = randomSeed + i
            continuousSet = continuous.SimulatedContinuousSet(
                self, localId, seed)
            continuousSet.setReferenceSet(referenceSet)
            self.addContinuousSet(continuousSet)
        # RnaQuantificationSets
        for i in range(numRnaQuantSets):
            localId = 'simRqs{}'.format(i)
            rnaQuantSet = rnaQuantification.SimulatedRnaQuantificationSet(
                self, localId)
            rnaQuantSet.setReferenceSet(referenceSet)
            self.addRnaQuantificationSet(rnaQuantSet)



"""
Functionality common to cli modules
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import ga4gh.server

import ga4gh.schemas.protocol as protocol


def addVersionArgument(parser):
    # TODO argparse strips newlines from version output
    versionString = (
        "GA4GH Server Version {}\n"
        "(Protocol Version {})".format(
            ga4gh.server.__version__, protocol.version))
    parser.add_argument(
        "--version", version=versionString, action="version")


def addDisableUrllibWarningsArgument(parser):
    parser.add_argument(
        "--disable-urllib-warnings", default=False, action="store_true",
        help="Disable urllib3 warnings")



"""
Simple tool to download files
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import sys

import humanize

import requests.packages.urllib3
requests.packages.urllib3.disable_warnings()


class FileDownloader(object):
    """
    Base class for file downloaders of different protocols
    """
    defaultStream = sys.stdout

    def __init__(self, url, path, stream=defaultStream):
        self.url = url
        self.path = path
        self.basename = path
        self.basenameLength = len(self.basename)
        self.stream = stream
        self.bytesReceived = 0
        self.displayIndex = 0
        self.displayWindowSize = 20
        self.fileSize = None
        self.displayCounter = 0

    def _printStartDownloadMessage(self):
        self.stream.write("Downloading '{}' to '{}'\n".format(
            self.url, self.path))

    def _cleanUp(self):
        self.stream.write("\n")
        self.stream.flush()

    def _getFileNameDisplayString(self):
        if self.basenameLength <= self.displayWindowSize:
            return self.basename
        else:
            return self.basename  # TODO scrolling window here

    def _updateDisplay(self, modulo=1):
        self.displayCounter += 1
        if self.displayCounter % modulo != 0:
            return
        fileName = self._getFileNameDisplayString()
        if self.fileSize is None:
            displayString = "{}   bytes received: {}\r"
            bytesReceived = humanize.filesize.naturalsize(
                self.bytesReceived)
            self.stream.write(displayString.format(
                fileName, bytesReceived))
        else:
            # TODO contentlength seems to slightly under-report how many
            # bytes we have to download... hence the min functions
            percentage = min(self.bytesReceived / self.fileSize, 1)
            numerator = humanize.filesize.naturalsize(
                min(self.bytesReceived, self.fileSize))
            denominator = humanize.filesize.naturalsize(
                self.fileSize)
            displayString = "{}   {:<6.2%} ({:>9} / {:<9})\r"
            self.stream.write(displayString.format(
                fileName, percentage, numerator, denominator))
        self.stream.flush()


class HttpFileDownloader(FileDownloader):
    """
    Provides a wget-like file download and terminal display for HTTP
    """
    defaultChunkSize = 1048576  # 1MB

    def __init__(self, url, path, chunkSize=defaultChunkSize,
                 stream=FileDownloader.defaultStream):
        super(HttpFileDownloader, self).__init__(
            url, path, stream)
        self.chunkSize = chunkSize

    def download(self):
        self._printStartDownloadMessage()
        response = requests.get(self.url, stream=True)
        response.raise_for_status()
        try:
            contentLength = int(response.headers['content-length'])
            self.fileSize = contentLength
        except KeyError:
            # chunked transfer encoding
            pass
        with open(self.path, 'wb') as outputFile:
            for chunk in response.iter_content(chunk_size=self.chunkSize):
                self.bytesReceived += self.chunkSize
                self._updateDisplay()
                outputFile.write(chunk)
        self._cleanUp()



"""
Random data based tests for the interval iterator code.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest
import random

import ga4gh.server.paging as paging


def setUp():
    """
    Set the random seed to ensure that the tests in this module are
    reproducable.
    """
    random.seed(1)


def intervalsIntersect(a1, b1, a2, b2):
    """
    Returns True if the specified half-closed intervals [a1, b1)
    and [a2, b2) intersect.
    """
    # Intervals do not interset if [a1, b1) is wholly to the left
    # or right of [a2, b2).
    return not (b1 <= a2 or a1 >= b2)


class TestIntervalsIntersect(unittest.TestCase):
    """
    Tests to ensure our interval intersection code works.
    """
    def testEmptyIntervals(self):
        self.assertFalse(intervalsIntersect(0, 0, 0, 0))
        self.assertFalse(intervalsIntersect(1, 1, 1, 1))
        self.assertFalse(intervalsIntersect(1, 1, 1, 2))
        self.assertFalse(intervalsIntersect(1, 10, 1, 1))

    def testNonIntersecting(self):
        tests = [
            [0, 1, 1, 2],
            [0, 1, 2, 3],
            [2, 1, 0, 1],
            [2, 3, 0, 1]]
        for test in tests:
            self.assertFalse(intervalsIntersect(*test))

    def testIntersecting(self):
        tests = [
            [0, 1, 0, 1],
            [0, 1, 0, 3],
            [0, 3, 0, 1],
            [1, 10, 9, 10],
            [1, 10, 9, 12],
            [1, 10, 5, 6],
            [1, 10, 1, 2]]
        for test in tests:
            self.assertTrue(intervalsIntersect(*test))


def randomIntervals(start, end, numIntervals):
    """
    Returns a list of numIntervals half-closed intervals. Each element
    of the returned array is a tuple a, b such that a <= start < b <= end.
    """
    intervals = []
    for _ in range(numIntervals):
        a = random.randrange(start, end - 1)
        b = random.randrange(a + 1, end)
        assert start <= a < b <= end
        intervals.append((a, b))
    return intervals


class IntervalSet(object):
    """
    Class representing a set of half-closed intervals.
    """
    def __init__(self, start, end, intervals):
        self.start = start
        self.end = end
        self.intervals = sorted(intervals, key=lambda x: x[0])

    def get(self, start, end):
        """
        Returns an iterator over all intervals in this set that intersect
        with the specified interval.
        """
        for interval in self.intervals:
            if intervalsIntersect(start, end, interval[0], interval[1]):
                yield interval


class FakeRequest(object):
    """
    A class to stand in as a Request object.
    """
    def __init__(self, start, end, pageToken):
        self.page_token = pageToken
        self.start = start
        self.end = end


class TrivialIntervalIterator(paging.IntervalIterator):
    """
    The simplest possible instance of the interval iterator
    used to test the iteration code.
    """
    def __init__(self, intervalSet, start, end, pageToken=None):
        self.intervalSet = intervalSet
        request = FakeRequest(start, end, pageToken)
        super(TrivialIntervalIterator, self).__init__(request, None)

    def _getContainer(self):
        return None

    def _search(self, start, end):
        return self.intervalSet.get(start, end)

    def _getStart(self, interval):
        return interval[0]

    def _getEnd(self, interval):
        return interval[1]


class TestIntervalIterator(unittest.TestCase):
    """
    A class to systematically test the paging code over interval search
    by randomly generating densely packed interval data, and comparing the
    results for a range of page sizes.
    """
    def setUp(self):
        self.num_random_tests = 5
        self.testIntervalSets = []
        # Set up some interval sets that provoke specific issues
        intervals = [
            (0, 1), (1, 8), (2, 9), (4, 7), (4, 8), (5, 9), (6, 7), (6, 7),
            (7, 8), (8, 9)]
        self.testIntervalSets.append(IntervalSet(0, 10, intervals))
        # Make two interval sets, and merge them. We want to have patches
        # of intervals, but with queries happening in between them.
        intervals = randomIntervals(10, 20, 10) + randomIntervals(30, 40, 10)
        self.testIntervalSets.append(IntervalSet(0, 50, intervals))
        # Add other random interval sets
        self.testIntervalSets.append(
            IntervalSet(0, 10, randomIntervals(0, 10, 10)))
        self.testIntervalSets.append(
            IntervalSet(0, 10, randomIntervals(0, 10, 10)))
        self.testIntervalSets.append(
            IntervalSet(0, 100, randomIntervals(0, 100, 10)))
        self.testIntervalSets.append(
            IntervalSet(0, 100, randomIntervals(0, 100, 100)))

    def verifyInterval(self, intervalSet, start, end):
        """
        Verify that we can pick up iteration of the interval from
        anywhere by starting a new iterator from every point.
        """
        topIterator = list(TrivialIntervalIterator(intervalSet, start, end))
        allIntervals = list(intervalSet.get(start, end))
        topIntervals = []
        for topInterval, topPageToken in topIterator[:-1]:
            topIntervals.append(topInterval)
            self.assertIsNotNone(topPageToken)
            # We should be able to pick the iteration up from here and go
            # forward, getting the same set of intervals
            subIterator = TrivialIntervalIterator(
                intervalSet, start, end, topPageToken)
            subIntervals = list(topIntervals)
            for subInterval, subPageToken in subIterator:
                subIntervals.append(subInterval)
            self.assertEqual(allIntervals, subIntervals)
            self.assertIsNone(subPageToken)
        topInterval, topPageToken = topIterator[-1]
        self.assertIsNone(topPageToken)
        topIntervals.append(topInterval)
        self.assertEqual(allIntervals, topIntervals)

    def testFullInterval(self):
        for intervalSet in self.testIntervalSets:
            self.verifyInterval(
                intervalSet, intervalSet.start, intervalSet.end)

    def verifyEmptyInterval(self, intervalSet, start, end):
        """
        Verify that we correctly return an empty iterator.
        """
        iterator = TrivialIntervalIterator(intervalSet, start, end)
        self.assertIsNone(next(iterator, None))

    def testEmptyInterval(self):
        for intervalSet in self.testIntervalSets:
            for start, end in [(-1, -1), (intervalSet.end, intervalSet.end)]:
                self.verifyEmptyInterval(intervalSet, start, end)

    def testStartInGap(self):
        for intervalSet in self.testIntervalSets:
            starts = set(start for start, _ in intervalSet.intervals)
            gaps = set(range(intervalSet.start, intervalSet.end - 1)) - starts
            start = -1
            if len(gaps) > 0:
                start = list(gaps)[0]
            self.verifyInterval(intervalSet, start, intervalSet.end)

    def testOutsideRange(self):
        for intervalSet in self.testIntervalSets:
            self.verifyInterval(
                intervalSet, intervalSet.start - 1, intervalSet.end)
            self.verifyInterval(
                intervalSet, intervalSet.start, intervalSet.end + 1)
            self.verifyInterval(
                intervalSet, intervalSet.start - 1, intervalSet.end + 1)
            self.verifyInterval(
                intervalSet, intervalSet.start - 10, intervalSet.end + 10)

    def testRandomStarts(self):
        for intervalSet in self.testIntervalSets:
            for _ in range(self.num_random_tests):
                start = random.randrange(
                    intervalSet.start, intervalSet.end - 1)
                end = intervalSet.end
                intervals = list(intervalSet.get(start, end))
                if len(intervals) == 0:
                    self.verifyEmptyInterval(intervalSet, start, end)
                else:
                    self.verifyInterval(intervalSet, start, end)

    def testRandomEnds(self):
        for intervalSet in self.testIntervalSets:
            for _ in range(self.num_random_tests):
                start = intervalSet.start
                end = random.randrange(
                    intervalSet.start + 1, intervalSet.end)
                intervals = list(intervalSet.get(start, end))
                if len(intervals) == 0:
                    self.verifyEmptyInterval(intervalSet, start, end)
                else:
                    self.verifyInterval(intervalSet, start, end)

    def testRandomIntervals(self):
        for intervalSet in self.testIntervalSets:
            for _ in range(self.num_random_tests):
                start = random.randrange(
                    intervalSet.start, intervalSet.end - 1)
                end = random.randrange(start, intervalSet.end - 1)
                intervals = list(intervalSet.get(start, end))
                if len(intervals) == 0:
                    self.verifyEmptyInterval(intervalSet, start, end)
                else:
                    self.verifyInterval(intervalSet, start, end)



"""
Helps to implement authentication and authorization using Auth0.

Offers functions for generating the view functions needed to implement Auth0,
a login screen, callback maker, and a function decorator for protecting
endpoints.
"""
import flask
import requests
import functools
import json

import jwt

import ga4gh.server.exceptions as exceptions


def auth_decorator(app=None):
    """
    This decorator wraps a view function so that it is protected when Auth0
    is enabled. This means that any request will be expected to have a signed
    token in the authorization header if the `AUTH0_ENABLED` configuration
    setting is True.

    The authorization header will have the form:

    "authorization: Bearer eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9....."

    If a request is not properly signed, an attempt is made to provide the
    client with useful error messages. This means that if a request is not
    authorized the underlying view function will not be executed.

    When `AUTH0_ENABLED` is false, this decorator will simply execute the
    decorated view without observing the authorization header.
    :param app:
    :return: Flask view decorator
    """
    def requires_auth(f):
        @functools.wraps(f)
        def decorated(*args, **kwargs):
            # This decorator will only apply with AUTH0_ENABLED set to True.
            if app.config.get('AUTH0_ENABLED', False):
                client_id = app.config.get("AUTH0_CLIENT_ID")
                client_secret = app.config.get("AUTH0_CLIENT_SECRET")
                auth_header = flask.request.headers.get('Authorization', None)
                # Each of these functions will throw a 401 is there is a
                # problem decoding the token with some helpful error message.
                if auth_header:
                    token, profile = decode_header(
                        auth_header, client_id, client_secret)
                else:
                    raise exceptions.NotAuthorizedException()
                # We store the token in the session so that later
                # stages can use it to connect identity and authorization.
                flask.session['auth0_key'] = token
                # Now we need to make sure that on top of having a good token
                # They are authorized, and if not provide an error message
                is_authorized(app.cache, profile['email'])
                is_active(app.cache, token)
            return f(*args, **kwargs)
        return decorated
    return requires_auth


def decode_header(auth_header, client_id, client_secret):
    """
    A function that threads the header through decoding and returns a tuple
    of the token and payload if successful. This does not fully authenticate
    a request.
    :param auth_header:
    :param client_id:
    :param client_secret:
    :return: (token, profile)
    """
    return _decode_header(
        _well_formed(
            _has_token(_has_bearer(_has_header(auth_header)))),
        client_id, client_secret)


def logout(cache):
    """
    Logs out the current session by removing it from the cache. This is
    expected to only occur when a session has
    """
    cache.set(flask.session['auth0_key'], None)
    flask.session.clear()
    return True


def callback_maker(
        cache=None, domain='', client_id='',
        client_secret='', redirect_uri=''):
    """
    This function will generate a view function that can be used to handle
    the return from Auth0. The "callback" is a redirected session from auth0
    that includes the token we can use to authenticate that session.

    If the session is properly authenticated Auth0 will provide a code so our
    application can identify the session. Once this has been done we ask
    for more information about the identified session from Auth0. We then use
    the email of the user logged in to Auth0 to authorize their token to make
    further requests by adding it to the application's cache.

    It sets a value in the cache that sets the current session as logged in. We
    can then refer to this id_token to later authenticate a session.

    :param domain:
    :param client_id:
    :param client_secret:
    :param redirect_uri:
    :return : View function
    """
    def callback_handling():
        code = flask.request.args.get('code')
        if code is None:
            raise exceptions.NotAuthorizedException(
                'The callback expects a well '
                'formatted code, {} was provided'.format(code))
        json_header = {'content-type': 'application/json'}
        # Get auth token
        token_url = "https://{domain}/oauth/token".format(domain=domain)
        token_payload = {
            'client_id':     client_id,
            'client_secret': client_secret,
            'redirect_uri':  redirect_uri,
            'code':          code,
            'grant_type':    'authorization_code'}
        try:
            token_info = requests.post(
                token_url,
                data=json.dumps(token_payload),
                headers=json_header).json()
            id_token = token_info['id_token']
            access_token = token_info['access_token']
        except Exception as e:
            raise exceptions.NotAuthorizedException(
                'The callback from Auth0 did not'
                'include the expected tokens: \n'
                '{}'.format(e.message))
        # Get profile information
        try:
            user_url = \
              "https://{domain}/userinfo?access_token={access_token}".format(
                  domain=domain, access_token=access_token)
            user_info = requests.get(user_url).json()
            email = user_info['email']
        except Exception as e:
            raise exceptions.NotAuthorizedException(
                'The user profile from Auth0 did '
                'not contain the expected data: \n {}'.format(e.message))
        # Log token in
        user = cache.get(email)
        if user and user['authorized']:
            cache.set(id_token, user_info)
            return flask.redirect('/login?code={}'.format(id_token))
        else:
            return flask.redirect('/login')
    return callback_handling


def render_login(
        app=None, scopes='', redirect_uri='', domain='', client_id=''):
    """
    This function will generate a view function that can be used to handle
    the return from Auth0. The "callback" is a redirected session from auth0
    that includes the token we can use to authenticate that session.

    If the session is properly authenticated Auth0 will provide a code so our
    application can identify the session. Once this has been done we ask
    for more information about the identified session from Auth0. We then use
    the email of the user logged in to Auth0 to authorize their token to make
    further requests by adding it to the application's cache.

    It sets a value in the cache that sets the current session as logged in. We
    can then refer to this id_token to later authenticate a session.

    :param app:
    :param scopes:
    :param redirect_uri:
    :param domain:
    :param client_id:
    :return : Rendered login template
    """
    return app.jinja_env.from_string(LOGIN_HTML).render(
        scopes=scopes,
        redirect_uri=redirect_uri,
        domain=domain,
        client_id=client_id)


def render_key(app, key=""):
    """
    Renders a view from the app and a key that lets the current session grab
    its token.
    :param app:
    :param key:
    :return: Rendered view
    """
    return app.jinja_env.from_string(KEY_HTML).render(
        key=key)


def authorize_email(email='davidcs@ucsc.edu', cache=None):
    """
    Adds an email address to the list of authorized emails stored in an
    ephemeral cache.
    :param email:
    """
    # TODO safely access cache
    cache.set(email, {'authorized': True})


def _has_header(auth_header):
    if not auth_header:
        raise exceptions.NotAuthorizedException(
            'Authorization header is expected.')
    return auth_header


def _has_bearer(auth_header):
    parts = auth_header.split()
    if parts[0].lower() != 'bearer':
        raise exceptions.NotAuthorizedException(
            'Authorization header must start with "Bearer".')
    return auth_header


def _has_token(auth_header):
    parts = auth_header.split()
    if len(parts) == 1:
        raise exceptions.NotAuthorizedException(
           'Token not found in header.')
    return auth_header


def _well_formed(auth_header):
    parts = auth_header.split()
    if len(parts) > 2:
        raise exceptions.NotAuthorizedException(
            'Authorization header must be Bearer + \s + token.')
    return auth_header


def _decode_header(auth_header, client_id, client_secret):
    """
    Takes the header and tries to return an active token and decoded
    payload.
    :param auth_header:
    :param client_id:
    :param client_secret:
    :return: (token, profile)
    """
    try:
        token = auth_header.split()[1]
        payload = jwt.decode(
            token,
            client_secret,
            audience=client_id)
    except jwt.ExpiredSignature:
        raise exceptions.NotAuthorizedException(
            'Token has expired, please log in again.')
    # is valid client
    except jwt.InvalidAudienceError:
        message = 'Incorrect audience, expected: {}'.format(
            client_id)
        raise exceptions.NotAuthorizedException(message)
    # is valid token
    except jwt.DecodeError:
        raise exceptions.NotAuthorizedException(
            'Token signature could not be validated.')
    except Exception as e:
        raise exceptions.NotAuthorizedException(
            'Token signature was malformed. {}'.format(e.message))
    return token, payload


def is_authorized(cache, email):
    if not cache.get(email):
        message = '{} is not authorized to ' \
                  'access this resource'.format(email)
        raise exceptions.NotAuthenticatedException(message)
    return email


def is_active(cache, token):
    """
    Accepts the cache and ID token and checks to see if the profile is
    currently logged in. If so, return the token, otherwise throw a
    NotAuthenticatedException.
    :param cache:
    :param token:
    :return:
    """
    profile = cache.get(token)
    if not profile:
        raise exceptions.NotAuthenticatedException(
            'The token is good, but you are not logged in. Please '
            'try logging in again.')
    return profile


# This HTML string is used to render the login page. It is a jinja template.
LOGIN_HTML = """<html>
<head>

  <title>Log in</title></head><body><div>
    <script src="https://cdn.auth0.com/js/lock/10.0/lock.min.js"></script>
    <script type="text/javascript">
      var lock = new Auth0Lock('{{ client_id }}', '{{ domain }}', {
        auth: {
          redirectUrl: '{{ redirect_uri }}',
          responseType: 'code',
          params: {
            scope: '{{ scopes }}' // https://auth0.com/docs/scopes
          }
        }
      });
    lock.show();
    </script>
    </div>"""

KEY_HTML = """<html>
<head>
<title>GA4GH Server API Token</title></head><body><div>
<h1>Your API Token</h1>
<p>Your token is now active, add it as your "Authorization: bearer $TOKEN" header
when making requests to protected endpoints</p>
<textarea cols=120 rows=5 onClick='this.select()' readonly>{{ key }}</textarea>
<h3><a href="/?key={{ key }}">Visit landing page</a></h3>
</div>
""" # noqa



"""
Tests the pysam sanitizer
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.server.datamodel as datamodel
import ga4gh.server.exceptions as exceptions


class TestPysamSanitizer(datamodel.PysamDatamodelMixin, unittest.TestCase):
    """
    Test the pysam sanitizer
    """
    def testSanitizeString(self):
        # only accept strings as params
        with self.assertRaises(exceptions.DatamodelValidationException):
            self.sanitizeString(1, 'longString')

        # return same string if string is short enough
        shortString = 'shortString'
        self.assertIsInstance(shortString, unicode)
        result = self.sanitizeString(shortString, shortString)
        self.assertEqual(shortString, result)
        self.assertNotIsInstance(result, unicode)

        # shorten string length if string too long
        longString = 'x' * (self.maxStringLength + 1)
        result = self.sanitizeString(longString, 'longString')
        self.assertEqual(longString[:self.maxStringLength], result)

    def testSanitizeInt(self):
        # only accept ints as params
        with self.assertRaises(exceptions.DatamodelValidationException):
            self.sanitizeInt('whoops', 0, 0, 0)

        # set to minVal if less than minVal
        minVal = 0
        result = self.sanitizeInt(-10, minVal, 100, 'int')
        self.assertEqual(minVal, result)

        # set to maxVal if more than maxVal
        maxVal = 100
        result = self.sanitizeInt(200, 0, maxVal, 'int')
        self.assertEqual(maxVal, result)

        # don't change if within range
        val = 50
        result = self.sanitizeInt(val, 0, 100, 'int')
        self.assertEqual(result, val)

    def testAssertValidRange(self):
        # wrong range should throw error
        with self.assertRaises(exceptions.DatamodelValidationException):
            self.assertValidRange(100, 0, 'start', 'end')

        # correct range should not throw error
        self.assertValidRange(0, 100, 'start', 'end')

    def testAssertInRange(self):
        # too low
        with self.assertRaises(exceptions.DatamodelValidationException):
            self.assertInRange(-1, 0, 100, 'example')

        # too high
        with self.assertRaises(exceptions.DatamodelValidationException):
            self.assertInRange(101, 0, 100, 'example')

        # in range
        self.assertInRange(50, 0, 100, 'example')

    def testAssertInt(self):
        # is an int
        self.assertInt(5, 'example')

        # is not an int
        with self.assertRaises(exceptions.DatamodelValidationException):
            self.assertInt('5', 'example')

    def testSanitizeVariantFileFetch(self):
        contigArg = 'x' * (self.maxStringLength + 1)
        startArg = self.vcfMin - 1
        stopArg = self.vcfMax + 1
        contig, start, stop = self.sanitizeVariantFileFetch(
            contigArg, startArg, stopArg)
        self.assertEqual(contigArg[:self.maxStringLength], contig)
        self.assertEqual(start, self.vcfMin)
        self.assertEqual(stop, self.vcfMax)

    def testSanitizeAlignmentFileFetch(self):
        startArg = self.samMin - 1
        endArg = self.samMaxEnd + 1
        start, end = self.sanitizeAlignmentFileFetch(startArg, endArg)
        self.assertEqual(start, self.samMin)
        self.assertEqual(end, self.samMaxEnd)



"""
Biodata objects
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import datetime
import json

import ga4gh.server.datamodel as datamodel
import ga4gh.server.exceptions as exceptions

import ga4gh.schemas.protocol as protocol


class Biosample(datamodel.DatamodelObject):
    """
    This class represents an abstract Biosample object.
    It sets default values and getters, as well as the
    toProtocolElement function.
    """
    compoundIdClass = datamodel.BiosampleCompoundId

    def __init__(self, parentContainer, localId):
        super(Biosample, self).__init__(parentContainer, localId)
        self._created = datetime.datetime.now().isoformat()
        self._updated = datetime.datetime.now().isoformat()
        self._description = None
        self._disease = None
        self._name = localId
        self._individualId = None
        self._datasetId = parentContainer.getId()
        self._individualAgeAtCollection = None

    def toProtocolElement(self):
        disease = None
        if self.getDisease():
            disease = protocol.fromJson(
                json.dumps(self.getDisease()), protocol.OntologyTerm)
        individualAgeAtCollection = None
        if self.getIndividualAgeAtCollection():
            individualAgeAtCollection = protocol.fromJson(
                json.dumps(self.getIndividualAgeAtCollection()), protocol.Age)
        biosample = protocol.Biosample(
            dataset_id=self._datasetId,
            created=self.getCreated(),
            updated=self.getUpdated(),
            description=self.getDescription(),
            id=self.getId(),
            individual_id=self.getIndividualId(),
            name=self.getName(),
            disease=disease,
            individual_age_at_collection=individualAgeAtCollection)
        self.serializeAttributes(biosample)
        return biosample

    def populateFromJson(self, jsonString):
        try:
            parsed = protocol.fromJson(jsonString, protocol.Biosample)
        except:
            raise exceptions.InvalidJsonException(jsonString)
        self._created = parsed.created
        self._updated = parsed.updated
        self._description = parsed.description
        self._disease = protocol.toJsonDict(parsed.disease)
        self._individualId = parsed.individual_id
        self._individualAgeAtCollection = protocol.toJsonDict(
                                           parsed.individual_age_at_collection)
        attributes = {}
        for key in parsed.attributes.attr:
            attributes[key] = {
                "values": protocol.toJsonDict(parsed.attributes.attr[key])}
        self.setAttributes(attributes)
        return self

    def populateFromRow(self, biosampleRecord):
        # TODO coerce to types
        self._created = biosampleRecord.created
        self._updated = biosampleRecord.updated
        self._description = biosampleRecord.description
        self._disease = json.loads(biosampleRecord.disease)
        self._individualId = biosampleRecord.individualid
        self.setAttributesJson(biosampleRecord.attributes)
        self._individualAgeAtCollection = json.loads(
                                biosampleRecord.individualAgeAtCollection)
        return self

    def setIndividualId(self, individualId):
        self._individualId = individualId

    def getIndividualId(self):
        return self._individualId

    def getCreated(self):
        return self._created

    def getUpdated(self):
        return self._updated

    def getDescription(self):
        return self._description

    def getDisease(self):
        if self._disease is not {}:
            return self._disease
        else:
            return None

    def getName(self):
        return self._name

    def getIndividualAgeAtCollection(self):
        if self._individualAgeAtCollection is not {}:
            return self._individualAgeAtCollection
        else:
            return None


class Individual(datamodel.DatamodelObject):
    """
    This class represents an abstract Individual object.
    It sets default values and getters, as well as the
    toProtocolElement function.
    """
    compoundIdClass = datamodel.IndividualCompoundId

    def __init__(self, parentContainer, localId):
        super(Individual, self).__init__(parentContainer, localId)
        self._created = datetime.datetime.now().isoformat()
        self._updated = datetime.datetime.now().isoformat()
        self._description = None
        self._species = None
        self._sex = None
        self._name = localId
        self._datasetId = parentContainer.getId()

    def toProtocolElement(self):
        species = None
        sex = None
        if self.getSpecies():
            species = protocol.fromJson(
                json.dumps(self.getSpecies()), protocol.OntologyTerm)
        if self.getSex():
            sex = protocol.fromJson(
                json.dumps(self.getSex()), protocol.OntologyTerm)
        gaIndividual = protocol.Individual(
            dataset_id=self._datasetId,
            created=self.getCreated(),
            updated=self.getUpdated(),
            description=self.getDescription(),
            id=self.getId(),
            name=self.getName(),
            species=species,
            sex=sex)
        self.serializeAttributes(gaIndividual)
        return gaIndividual

    def populateFromRow(self, individualRecord):
        # TODO coerce to types
        self._name = individualRecord.name
        self._created = individualRecord.created
        self._updated = individualRecord.updated
        self._description = individualRecord.description
        self._species = json.loads(individualRecord.species)
        self._sex = json.loads(individualRecord.sex)
        self.setAttributesJson(individualRecord.attributes)
        return self

    def populateFromJson(self, jsonString):
        # TODO validate
        try:
            parsed = protocol.fromJson(jsonString, protocol.Individual)
        except:
            raise exceptions.InvalidJsonException(jsonString)
        self._created = parsed.created
        self._updated = parsed.updated
        self._description = parsed.description
        self._species = protocol.toJsonDict(parsed.species)
        self._sex = protocol.toJsonDict(parsed.sex)
        attributes = {}
        for key in parsed.attributes.attr:
            attributes[key] = {
                "values": protocol.toJsonDict(parsed.attributes.attr[key])}
        self.setAttributes(attributes)
        return self

    def getCreated(self):
        return self._created

    def getUpdated(self):
        return self._updated

    def getDescription(self):
        return self._description

    def getSpecies(self):
        return self._species

    def getSex(self):
        return self._sex

    def getName(self):
        return self._name



"""
Constructs a data source for the ga4gh server by downloading data from
authoritative remote servers.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import gzip
import os
import requests
import shutil
import subprocess
import tempfile
import urllib2

import pysam

import ga4gh.common.utils as utils
import glue

glue.ga4ghImportGlue()

# We need to turn off QA because of the import glue
import ga4gh.server.datarepo as datarepo  # NOQA
import ga4gh.server.datamodel.references as references  # NOQA
import ga4gh.server.datamodel.datasets as datasets  # NOQA
import ga4gh.server.datamodel.variants as variants  # NOQA
import ga4gh.server.datamodel.reads as reads  # NOQA


class ChromMinMax(object):
    """
    A container class for storing the min and max position seen
    for every chromosome
    """
    defaultMinPos = 2**30
    defaultMaxPos = 0

    class MinMax(object):
        def __init__(self):
            self.minPos = ChromMinMax.defaultMinPos
            self.maxPos = ChromMinMax.defaultMaxPos

    def __init__(self):
        self.chromMap = {}

    def addPos(self, chrom, position):
        if chrom not in self.chromMap:
            self.chromMap[chrom] = self.MinMax()
        minMax = self.chromMap[chrom]
        if position > minMax.maxPos:
            minMax.maxPos = position
        if position < minMax.minPos:
            minMax.minPos = position

    def getMinPos(self, chrom):
        minMax = self.chromMap[chrom]
        return minMax.minPos

    def getMaxPos(self, chrom):
        minMax = self.chromMap[chrom]
        return minMax.maxPos


def _fetchSequence(ac, startIndex=None, endIndex=None):
    """Fetch sequences from NCBI using the eself interface.

    An interbase interval may be optionally provided with startIndex and
    endIndex. NCBI eself will return just the requested subsequence, which
    might greatly reduce payload sizes (especially with chromosome-scale
    sequences). When wrapped is True, return list of sequence lines rather
    than concatenated sequence.

    >>> len(_fetchSequence('NP_056374.2'))
    1596

    Pass the desired interval rather than using Python's [] slice
    operator.

    >>> _fetchSequence('NP_056374.2',0,10)
    'MESRETLSSS'

    >>> _fetchSequence('NP_056374.2')[0:10]
    'MESRETLSSS'

    """
    urlFmt = (
        "http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?"
        "db=nucleotide&id={ac}&rettype=fasta&retmode=text")
    if startIndex is None or endIndex is None:
        url = urlFmt.format(ac=ac)
    else:
        urlFmt += "&seq_start={start}&seq_stop={stop}"
        url = urlFmt.format(ac=ac, start=startIndex + 1, stop=endIndex)
    resp = requests.get(url)
    resp.raise_for_status()
    seqlines = resp.content.splitlines()[1:]
    print("{ac}[{s},{e}) => {n} lines ({u})".format(
        ac=ac, s=startIndex, e=endIndex, n=len(seqlines), u=url))
    # return response as list of lines, already line wrapped
    return seqlines


class AbstractFileDownloader(object):
    """
    Base class for individual site genome file downloaders
    """
    def __init__(self, args):
        self.excludeReferenceMin = args.exclude_reference_min
        self.maxVariants = args.num_variants
        self.maxReads = args.num_reads
        self.samples = args.samples.split(',')
        self.tempDir = tempfile.mkdtemp(prefix="ga4gh-download")
        self.numChromosomes = args.num_chromosomes
        self.chromosomes = [str(j + 1) for j in range(self.numChromosomes)]
        self.dirName = args.destination
        self.datasetName = '1kg-p3-subset'
        self.variantSetName = 'mvncall'
        self.referenceSetName = 'GRCh37-subset'
        self.chromMinMax = ChromMinMax()
        self.accessions = {
            '1': 'CM000663.1',
            '2': 'CM000664.1',
            '3': 'CM000665.1',
            '4': 'CM000666.1',
            '5': 'CM000667.1',
            '6': 'CM000668.1',
            '7': 'CM000669.1',
            '8': 'CM000670.1',
            '9': 'CM000671.1',
            '10': 'CM000672.1',
            '11': 'CM000673.1',
            '12': 'CM000674.1',
            '13': 'CM000675.1',
            '14': 'CM000676.1',
            '15': 'CM000677.1',
            '16': 'CM000678.1',
            '17': 'CM000679.1',
            '18': 'CM000680.1',
            '19': 'CM000681.1',
            '20': 'CM000682.1',
            '21': 'CM000683.1',
            '22': 'CM000684.1',
        }
        self.studyMap = {
            'HG00096': 'GBR',
            'HG00533': 'CHS',
            'HG00534': 'CHS',
        }
        self.vcfFilePaths = []
        self.bamFilePaths = []
        self.fastaFilePath = None
        if os.path.exists(self.dirName) and args.force:
            shutil.rmtree(self.dirName)
        os.mkdir(self.dirName)
        self.repoPath = os.path.join(self.dirName, "registry.db")

    def log(self, message):
        print(message)

    def runCommand(self, command):
        """
        Runs the specified command.
        """
        subprocess.check_call(command, shell=True)

    def getVcfBaseUrl(self):
        return os.path.join(self.getBaseUrl(), 'ftp/release/20130502/')

    def getBamBaseUrl(self):
        return os.path.join(self.getBaseUrl(), 'ftp/phase3/data/')

    def _updatePositions(self, fileName):
        localVariantFile = pysam.VariantFile(fileName)
        localIterator = localVariantFile.fetch()
        for record in localIterator:
            self.chromMinMax.addPos(record.chrom, record.start)
        localIterator = None
        localVariantFile.close()
        self.log('chrom: {}, maxPos: {}, minPos: {}'.format(
            record.chrom, self.chromMinMax.getMaxPos(record.chrom),
            self.chromMinMax.getMinPos(record.chrom)))

    def _writeVcfTempFile(self, localTempFileName, data):
        with tempfile.NamedTemporaryFile() as binaryFile:
            binaryFile.write(data)
            binaryFile.flush()
            gzipFile = gzip.open(binaryFile.name, "r")
            outputFile = open(localTempFileName, "w")
            lineCount = 0
            for line in gzipFile:
                outputFile.write(line)
                if not line.startswith("#"):
                    lineCount += 1
                if lineCount >= self.maxVariants:
                    break
            assert lineCount == self.maxVariants
            outputFile.close()
            gzipFile.close()

    def _downloadVcf(self, chromosome):
        sourceFileName = (
            "ALL.chr{}.phase3_shapeit2_mvncall_integrated_v5a"
            ".20130502.genotypes.vcf.gz").format(chromosome)
        url = os.path.join(self.getVcfBaseUrl(), sourceFileName)
        self.log("Downloading '{}'".format(url))
        response = urllib2.urlopen(url)
        megabyte = 1024 * 1024
        data = response.read(megabyte)
        localFileName = os.path.join(
            self.dirName, "chr{}.vcf".format(chromosome))
        localCompressedFileName = "{}.gz".format(localFileName)
        localTempFileName = localFileName + '.unsampled'
        self.log("Writing '{}'".format(localTempFileName))
        self._writeVcfTempFile(localTempFileName, data)
        self.log("Sampling '{}'".format(localTempFileName))
        self.runCommand(
            'bcftools view --force-samples -s {} {} -o {}'.format(
                args.samples, localTempFileName, localFileName))
        os.remove(localTempFileName)
        self.log("Compressing '{}'".format(localFileName))
        self.runCommand('bgzip -f {}'.format(localFileName))
        self.log("Indexing '{}'".format(localCompressedFileName))
        self.runCommand('tabix {}'.format(localCompressedFileName))
        self._updatePositions(localCompressedFileName)
        self.vcfFilePaths.append(
            (localCompressedFileName, localCompressedFileName + ".tbi"))

    def downloadVcfs(self):
        for chromosome in self.chromosomes:
            self._downloadVcf(chromosome)

    def createBamHeader(self, baseHeader):
        """
        Creates a new bam header based on the specified header from the
        parent BAM file.
        """
        header = dict(baseHeader)
        newSequences = []
        for index, referenceInfo in enumerate(header['SQ']):
            if index < self.numChromosomes:
                referenceName = referenceInfo['SN']
                # The sequence dictionary in the BAM file has to match up
                # with the sequence ids in the data, so we must be sure
                # that these still match up.
                assert referenceName == self.chromosomes[index]
                newReferenceInfo = {
                    'AS': self.referenceSetName,
                    'SN': referenceName,
                    'LN': 0,  # FIXME
                    'UR': 'http://example.com',
                    'M5': 'dbb6e8ece0b5de29da56601613007c2a',  # FIXME
                    'SP': 'Human'
                }
                newSequences.append(newReferenceInfo)
        header['SQ'] = newSequences
        return header

    def _downloadIndex(self, indexUrl, localIndexFile):
        self.log("Downloading index from {} to {}".format(
            indexUrl, localIndexFile))
        response = urllib2.urlopen(indexUrl)
        with open(localIndexFile, "w") as destFile:
            destFile.write(response.read())

    def _downloadBam(self, sample):
        samplePath = '{}/alignment/'.format(sample)
        study = self.studyMap[sample]
        sourceFileName = (
            '{}.mapped.ILLUMINA.bwa.{}.'
            'low_coverage.20120522.bam'.format(sample, study))
        destFileName = os.path.join(
            self.dirName, "{}.bam".format(sample))
        baseUrl = self.getBamBaseUrl()
        sampleUrl = os.path.join(baseUrl, samplePath, sourceFileName)
        indexUrl = sampleUrl + ".bai"
        localIndexFile = os.path.join(self.tempDir, sourceFileName + ".bai")
        self._downloadIndex(indexUrl, localIndexFile)
        remoteFile = pysam.AlignmentFile(
            sampleUrl, filepath_index=localIndexFile)
        header = self.createBamHeader(remoteFile.header)
        self.log("Writing '{}'".format(destFileName))
        localFile = pysam.AlignmentFile(
            destFileName, 'wb', header=header)
        for chromosome in self.chromosomes:
            self.log("chromosome {}".format(chromosome))
            iterator = remoteFile.fetch(
                chromosome.encode('utf-8'),
                start=self.chromMinMax.getMinPos(chromosome),
                end=self.chromMinMax.getMaxPos(chromosome))
            for index, record in enumerate(iterator):
                # We only write records where we have the references
                # for the next mate. TODO we should take the positions
                # of these reads into account later when calculating
                # our reference bounds.
                if record.next_reference_id < self.numChromosomes:
                    if index >= self.maxReads:
                        break
                    localFile.write(record)
            self.log("{} records written".format(index))
        remoteFile.close()
        localFile.close()
        self.log("Indexing '{}'".format(destFileName))
        pysam.index(destFileName.encode('utf-8'))
        self.bamFilePaths.append(
            (destFileName, destFileName + ".bai"))

    def downloadBams(self):
        for sample in self.samples:
            self._downloadBam(sample)

    def _downloadFasta(self, chromosomes):
        fileName = os.path.join(self.dirName, "GRCh37-subset.fa")
        with open(fileName, "w") as outFasta:
            for chromosome in chromosomes:
                accession = self.accessions[chromosome]
                minPos = 0
                if self.excludeReferenceMin:
                    minPos = self.chromMinMax.getMinPos(chromosome)
                maxPos = self.chromMinMax.getMaxPos(chromosome)
                print(minPos, maxPos)
                print(">{}".format(chromosome), file=outFasta)
                sequence = _fetchSequence(accession, minPos, maxPos)
                for line in sequence:
                    print(line, file=outFasta)
        self.log("Compressing {}".format(fileName))
        self.runCommand("bgzip -f {}".format(fileName))
        compressedFileName = fileName + '.gz'
        self.log("Indexing {}".format(compressedFileName))
        self.runCommand("samtools faidx {}".format(compressedFileName))
        self.fastaFilePath = compressedFileName

    def downloadReference(self):
        self._downloadFasta(self.chromosomes)

    def createRepo(self):
        """
        Creates the repository for all the data we've just downloaded.
        """
        repo = datarepo.SqlDataRepository(self.repoPath)
        repo.open("w")
        repo.initialise()

        referenceSet = references.HtslibReferenceSet("GRCh37-subset")
        referenceSet.populateFromFile(self.fastaFilePath)
        referenceSet.setDescription("Subset of GRCh37 used for demonstration")
        referenceSet.setSpeciesFromJson(
                '{"id": "9606",'
                + '"term": "Homo sapiens", "source_name": "NCBI"}')
        for reference in referenceSet.getReferences():
            reference.setSpeciesFromJson(
                '{"id": "9606",'
                + '"term": "Homo sapiens", "source_name": "NCBI"}')
            reference.setSourceAccessions(
                self.accessions[reference.getName()] + ".subset")
        repo.insertReferenceSet(referenceSet)

        dataset = datasets.Dataset("1kg-p3-subset")
        dataset.setDescription("Sample data from 1000 Genomes phase 3")
        repo.insertDataset(dataset)

        variantSet = variants.HtslibVariantSet(dataset, "mvncall")
        variantSet.setReferenceSet(referenceSet)
        dataUrls = [vcfFile for vcfFile, _ in self.vcfFilePaths]
        indexFiles = [indexFile for _, indexFile in self.vcfFilePaths]
        variantSet.populateFromFile(dataUrls, indexFiles)
        variantSet.checkConsistency()
        repo.insertVariantSet(variantSet)

        for sample, (bamFile, indexFile) in zip(
                self.samples, self.bamFilePaths):
            readGroupSet = reads.HtslibReadGroupSet(dataset, sample)
            readGroupSet.populateFromFile(bamFile, indexFile)
            readGroupSet.setReferenceSet(referenceSet)
            repo.insertReadGroupSet(readGroupSet)

        repo.commit()
        repo.close()
        self.log("Finished creating the repository; summary:\n")
        repo.open("r")
        repo.printSummary()

    def cleanup(self):
        self.log('Removing temporary files')
        shutil.rmtree(self.tempDir)


class NcbiFileDownloader(AbstractFileDownloader):
    """
    Downloads files from NCBI
    """
    def getBaseUrl(self):
        return 'ftp://ftp-trace.ncbi.nih.gov/1000genomes'


class EbiFileDownloader(AbstractFileDownloader):
    """
    Downloads files from EBI
    """
    def getBaseUrl(self):
        return 'ftp://ftp.1000genomes.ebi.ac.uk/vol1'


sources = {
    "ncbi": NcbiFileDownloader,
    "ebi": EbiFileDownloader,
}


def parseArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "destination",
        help="the name of the directory that the data is downloaded to")
    parser.add_argument(
        "-f", "--force", default=False, action="store_true",
        help="Overwrite an existing directory with the same name")
    parser.add_argument(
        "--source", default="ncbi", choices=sources.keys(),
        help="the source to download from")
    parser.add_argument(
        "--samples", default='HG00096,HG00533,HG00534',
        help="a comma-seperated list of samples to download")
    parser.add_argument(
        "--num-reads", "-r", default=1000, type=int,
        help="the number of reads to download per reference")
    parser.add_argument(
        "--num-variants", "-V", default=1000, type=int,
        help="the maximum number of variants to download per VCF file.")
    parser.add_argument(
        "--num-chromosomes", default="3", type=int,
        help=(
            "the number of chromosomes whose corresponding reads should "
            "be downloaded"))
    parser.add_argument(
        "--exclude-reference-min", default=False, action="store_true",
        help="Exclude bases in the reference before the minimum position")
    args = parser.parse_args()
    return args


@utils.Timed()
def main(args):
    requiredExecutables = ['bcftools', 'tabix', 'bgzip', 'samtools']
    utils.requireExecutables(requiredExecutables)
    downloaderClass = sources[args.source]
    downloader = downloaderClass(args)
    try:
        downloader.downloadVcfs()
        downloader.downloadReference()
        downloader.downloadBams()
        downloader.createRepo()
    finally:
        downloader.cleanup()


if __name__ == '__main__':
    args = parseArgs()
    main(args)



"""
Simple shim for running the configuration testing program during development.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import dev_glue  # NOQA
import ga4gh.server.cli.configtest as cli_configtest

if __name__ == "__main__":
    cli_configtest.configtest_main()



"""
Tests the cli
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest
import shlex

import ga4gh.server.cli.server as cli_server
import ga4gh.server.cli.repomanager as cli_repomanager

import ga4gh.schemas.protocol as protocol


class TestServerArguments(unittest.TestCase):
    """
    Tests that the server can parse expected arguments
    """
    def testParseArguments(self):
        cliInput = """--port 7777 --host 123.4.5.6 --config MockConfigName
        --config-file /path/to/config --tls --dont-use-reloader"""
        parser = cli_server.getServerParser()
        args = parser.parse_args(cliInput.split())
        self.assertEqual(args.port, 7777)
        self.assertEqual(args.host, "123.4.5.6")
        self.assertEqual(args.config, "MockConfigName")
        self.assertEqual(args.config_file, "/path/to/config")
        self.assertTrue(args.tls)
        self.assertTrue(args.dont_use_reloader)


class TestRepoManagerCli(unittest.TestCase):

    def setUp(self):
        self.parser = cli_repomanager.RepoManager.getParser()
        self.registryPath = 'a/repo/path'
        self.datasetName = "datasetName"
        self.filePath = 'a/file/path'
        self.dirPath = 'a/dir/path/'
        self.individualName = "test"
        self.biosampleName = "test"
        self.individual = protocol.toJson(
            protocol.Individual(
                name="test",
                created="2016-05-19T21:00:19Z",
                updated="2016-05-19T21:00:19Z"))
        self.biosample = protocol.toJson(
            protocol.Biosample(
                name="test",
                created="2016-05-19T21:00:19Z",
                updated="2016-05-19T21:00:19Z"))

    def testInit(self):
        cliInput = "init {}".format(self.registryPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.runner, "init")

    def testVerify(self):
        cliInput = "verify {}".format(self.registryPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.runner, "verify")

    def testList(self):
        cliInput = "list {}".format(self.registryPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.runner, "list")

    def testAddDataset(self):
        cliInput = "add-dataset {} {}".format(
            self.registryPath, self.datasetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.runner, "addDataset")

    def testRemoveDataset(self):
        cliInput = "remove-dataset {} {} -f".format(
            self.registryPath, self.datasetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.runner, "removeDataset")
        self.assertEquals(args.force, True)

    def testAddReferenceSet(self):
        description = "description"
        cliInput = (
            "add-referenceset {} {} --description={} "
            "--species NCBITAXONID-JSON "
            "--isDerived True "
            "--assemblyId ASSEMBLYID "
            "--sourceAccessions SOURCEACCESSIONS "
            "--sourceUri SOURCEURI ").format(
            self.registryPath, self.filePath, description)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.filePath, self.filePath)
        self.assertEquals(args.description, description)
        self.assertEquals(args.species, "NCBITAXONID-JSON")
        self.assertEquals(args.isDerived, True)
        self.assertEquals(args.assemblyId, "ASSEMBLYID")
        self.assertEquals(args.sourceAccessions, "SOURCEACCESSIONS")
        self.assertEquals(args.sourceUri, "SOURCEURI")
        self.assertEquals(args.runner, "addReferenceSet")

    def testRemoveReferenceSet(self):
        referenceSetName = "referenceSetName"
        cliInput = "remove-referenceset {} {} -f".format(
            self.registryPath, referenceSetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.referenceSetName, referenceSetName)
        self.assertEquals(args.runner, "removeReferenceSet")
        self.assertEquals(args.force, True)

    def testAddReadGroupSet(self):
        cliInput = "add-readgroupset {} {} {} ".format(
            self.registryPath, self.datasetName, self.filePath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dataFile, self.filePath)
        self.assertEquals(args.indexFile, None)
        self.assertEquals(args.runner, "addReadGroupSet")

    def testAddReadGroupSetWithIndexFile(self):
        indexPath = self.filePath + ".bai"
        cliInput = "add-readgroupset {} {} {} -I {}".format(
            self.registryPath, self.datasetName, self.filePath,
            indexPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dataFile, self.filePath)
        self.assertEquals(args.indexFile, indexPath)
        self.assertEquals(args.runner, "addReadGroupSet")

    def testRemoveReadGroupSet(self):
        readGroupSetName = "readGroupSetName"
        cliInput = "remove-readgroupset {} {} {} -f".format(
            self.registryPath, self.datasetName, readGroupSetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.readGroupSetName, readGroupSetName)
        self.assertEquals(args.runner, "removeReadGroupSet")
        self.assertEquals(args.force, True)

    def testAddVariantSet(self):
        cliInput = "add-variantset {} {} {} ".format(
            self.registryPath, self.datasetName, self.filePath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dataFiles, [self.filePath])
        self.assertEquals(args.indexFiles, None)
        self.assertEquals(args.runner, "addVariantSet")

    def testAddVariantSetWithIndexFiles(self):
        file1 = "file1"
        file2 = "file2"
        indexFile1 = file1 + ".tbi"
        indexFile2 = file2 + ".tbi"
        cliInput = "add-variantset {} {} {} {} -I {} {}".format(
            self.registryPath, self.datasetName, file1, file2,
            indexFile1, indexFile2)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dataFiles, [file1, file2])
        self.assertEquals(args.indexFiles, [indexFile1, indexFile2])
        self.assertEquals(args.runner, "addVariantSet")

    def testRemoveVariantSet(self):
        variantSetName = "variantSetName"
        cliInput = "remove-variantset {} {} {}".format(
            self.registryPath, self.datasetName, variantSetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.variantSetName, variantSetName)
        self.assertEquals(args.runner, "removeVariantSet")
        self.assertEquals(args.force, False)

    def testAddOntology(self):
        cliInput = "add-ontology {} {}".format(
            self.registryPath, self.filePath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.filePath, self.filePath)
        self.assertEquals(args.runner, "addOntology")

    def testRemoveOntology(self):
        ontologyName = "the_ontology_name"
        cliInput = "remove-ontology {} {}".format(
            self.registryPath, ontologyName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.ontologyName, ontologyName)
        self.assertEquals(args.runner, "removeOntology")
        self.assertEquals(args.force, False)

    def testAddBiosample(self):
        cliInput = "add-biosample {} {} {} '{}'".format(
            self.registryPath,
            self.datasetName,
            self.biosampleName,
            self.biosample)
        args = self.parser.parse_args(shlex.split(cliInput))
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.biosampleName, self.biosampleName)
        self.assertEquals(args.biosample, self.biosample)
        self.assertEquals(args.runner, "addBiosample")

    def testRemoveBiosample(self):
        cliInput = "remove-biosample {} {} {}".format(
            self.registryPath,
            self.datasetName,
            self.biosampleName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.biosampleName, self.biosampleName)
        self.assertEquals(args.runner, "removeBiosample")
        self.assertEquals(args.force, False)

    def testAddIndividual(self):
        cliInput = "add-individual {} {} {} '{}'".format(
            self.registryPath,
            self.datasetName,
            self.individualName,
            self.individual)
        args = self.parser.parse_args(shlex.split(cliInput))
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.individualName, self.individualName)
        self.assertEquals(args.individual, self.individual)
        self.assertEquals(args.runner, "addIndividual")

    def testRemoveIndividual(self):
        cliInput = "remove-individual {} {} {}".format(
            self.registryPath,
            self.datasetName,
            self.individualName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.individualName, self.individualName)
        self.assertEquals(args.runner, "removeIndividual")
        self.assertEquals(args.force, False)

    def testAddPhenotypeAssociationSet(self):
        cliInput = "add-phenotypeassociationset {} {} {} -n NAME".format(
            self.registryPath,
            self.datasetName,
            self.dirPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dirPath, self.dirPath)
        self.assertEquals(args.name, "NAME")

    def testRemovePhenotypeAssociationSet(self):
        cliInput = "remove-phenotypeassociationset {} {} NAME".format(
            self.registryPath, self.datasetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.name, "NAME")



"""
Module responsible for translating continuous sequence annotation data
into GA4GH native objects.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import random
import re
import math

# no step/span; requires numpy
import pyBigWig

# for running bigwig tool externally
import subprocess

import ga4gh.server.datamodel as datamodel
import ga4gh.server.exceptions as exceptions
import ga4gh.schemas.pb as pb
import ga4gh.schemas.protocol as protocol


"""
These classes handle continuous message.
This message defines a format for exchanging continuous valued signal data,
such as those produced experimentally (e.g. ChIP-Seq data) or through
calculations (e.g. conservation scores). It can be used, for example,
to share data stored in Wiggle, BigWig, and BedGraph formats.

Assumes 0-based for everything (wiggle is 1-based).

The pyBiWig package is used to read the bigwig file it doesn't return step
and span, though. Other packages were
considered. The bx-python package is missing the bigwig reader in the
pypi verions. The ngslib package is not being maintained is difficult
to install.

"""


class WiggleReader:
    """
    Class for reading Wiggle data (from a file or a pipe) and returning
    protocol objects within the defined query region.

    Currently, only a single protocol object is returned, using getData().
    Unsampled values are assigned NaN in the object values array.
    The object start is the position of the first NaN value in the
    the sampled region. The values array extends to the last value
    that is not NaN. So, rather than covering the query region, the
    values array drops any initial NaN or final NaN. This was done
    to shorten the message and to avoid sending values containing only NaN.
    """

    # Wiggle has two modes:
    # 1. variable step, where each data line has a position and value.
    # 2. fixed step, where each data line has only a value. Positions
    #    are calculated as a fixed length from a start position, which
    #    are given in the header line (i.e. 'fixedStep ...').
    _VARIABLE_STEP = 1
    _FIXED_STEP = 2

    def __init__(self, reference, start, end):
        self._queryReference = reference
        self._queryStart = start
        self._queryEnd = end
        self._mode = None
        self._span = 1
        self._step = 1
        self._data = protocol.Continuous()
        self._position = None

    def getData(self):
        return self._data

    def parseStep(self, line):
        """
        Parse the line describing the mode.

        One of:
        variableStep chrom=<reference> [span=<window_size>]
        fixedStep chrom=<reference> start=<position> step=<step_interval>
                  [span=<window_size>]

        Span is optional, defaulting to 1. It indicates that each value
        applies to region, starting at the given position and extending
        <span> positions.
        """
        fields = dict([field.split('=') for field in line.split()[1:]])

        if 'chrom' in fields:
            self._reference = fields['chrom']
        else:
            raise ValueError("Missing chrom field in %s" % line.strip())

        if line.startswith("fixedStep"):
            if 'start' in fields:
                self._start = int(fields['start']) - 1  # to 0-based
            else:
                raise ValueError("Missing start field in %s" % line.strip())

        if 'span' in fields:
            self._span = int(fields['span'])
        if 'step' in fields:
            self._step = int(fields['step'])

    def readWiggleLine(self, line):
        """
        Read a wiggle line. If it is a data line, add values to the
        protocol object.
        """
        if(line.isspace() or line.startswith("#")
                or line.startswith("browser") or line.startswith("track")):
            return
        elif line.startswith("variableStep"):
            self._mode = self._VARIABLE_STEP
            self.parseStep(line)
            return
        elif line.startswith("fixedStep"):
            self._mode = self._FIXED_STEP
            self.parseStep(line)
            return
        elif self._mode is None:
            raise ValueError("Unexpected input line: %s" % line.strip())

        if self._queryReference != self._reference:
            return

        # read data lines
        fields = line.split()
        if self._mode == self._VARIABLE_STEP:
            start = int(fields[0])-1  # to 0-based
            val = float(fields[1])
        else:
            start = self._start
            self._start += self._step
            val = float(fields[0])

        if start < self._queryEnd and start > self._queryStart:
            if self._position is None:
                self._position = start
                self._data.start = start

            # fill gap
            while self._position < start:
                self._data.values.append(float('NaN'))
                self._position += 1
            for _ in xrange(self._span):
                self._data.values.append(val)
            self._position += self._span

    def wiggleFileHandleToProtocol(self, fileHandle):
        """
        Return a continuous protocol object satsifiying the given query
        parameters from the given wiggle file handle.
        """
        for line in fileHandle:
            self.readWiggleLine(line)
        return self._data

    def wiggleFileToProtocol(self, fileName):
        """
        Return a continuous protocol object satsifiying the given query
        parameters from the given wiggle file.
        """
        with open(fileName, 'r') as f:
            return self.wiggleFileHandleToProtocol(f)


class BigWigDataSource:
    """
    Class for reading from bigwig files.

    Two different readers are implemented:
    1. pyBigWig: a python package that wraps custom C code. It
       doesn't seem to return span and step values, limiting it's use.
    2. bigWigToWig: This is a command line tool for the Kent library.
       It must be installed separately.
    """

    def __init__(self, sourceFile):
        self._sourceFile = sourceFile
        self._INCREMENT = 10000  # max results per bw query
        self._MAX_VALUES = 1000  # max values length

    def checkReference(self, reference):
        """
        Check the reference for security. Tries to avoid any characters
        necessary for doing a script injection.
        """
        pattern = re.compile(r'[\s,;"\'&\\]')
        if pattern.findall(reference.strip()):
            return False
        return True

    def readValuesPyBigWig(self, reference, start, end):
        """
        Use pyBigWig package to read a BigWig file for the
        given range and return a protocol object.

        pyBigWig returns an array of values that fill the query range.
        Not sure if it is possible to get the step and span.

        This method trims NaN values from the start and end.

        pyBigWig throws an exception if end is outside of the
        reference range. This function checks the query range
        and throws its own exceptions to avoid the ones thrown
        by pyBigWig.
        """
        if not self.checkReference(reference):
            raise exceptions.ReferenceNameNotFoundException(reference)
        if start < 0:
            start = 0
        bw = pyBigWig.open(self._sourceFile)
        referenceLen = bw.chroms(reference)
        if referenceLen is None:
            raise exceptions.ReferenceNameNotFoundException(reference)
        if end > referenceLen:
            end = referenceLen
        if start >= end:
            raise exceptions.ReferenceRangeErrorException(
                reference, start, end)

        data = protocol.Continuous()
        curStart = start
        curEnd = curStart + self._INCREMENT
        while curStart < end:
            if curEnd > end:
                curEnd = end
            for i, val in enumerate(bw.values(reference, curStart, curEnd)):
                if not math.isnan(val):
                    if len(data.values) == 0:
                        data.start = curStart + i
                    data.values.append(val)
                    if len(data.values) == self._MAX_VALUES:
                        yield data
                        data = protocol.Continuous()
                elif len(data.values) > 0:
                    # data.values.append(float('NaN'))
                    yield data
                    data = protocol.Continuous()
            curStart = curEnd
            curEnd = curStart + self._INCREMENT

        bw.close()
        if len(data.values) > 0:
            yield data

    def readValuesBigWigToWig(self, reference, start, end):
        """
        Read a bigwig file and return a protocol object with values
        within the query range.

        This method uses the bigWigToWig command line tool from UCSC
        GoldenPath. The tool is used to return values within a query region.
        The output is in wiggle format, which is processed by the WiggleReader
        class.

        There could be memory issues if the returned results are large.

        The input reference can be a security problem (script injection).
        Ideally, it should be checked against a list of known chromosomes.
        Start and end should not be problems since they are integers.
        """
        if not self.checkReference(reference):
            raise exceptions.ReferenceNameNotFoundException(reference)
        if start < 0:
            raise exceptions.ReferenceRangeErrorException(
                reference, start, end)
            # TODO: CHECK IF QUERY IS BEYOND END

        cmd = ["bigWigToWig", self._sourceFile, "stdout", "-chrom="+reference,
               "-start="+str(start), "-end="+str(end)]
        wiggleReader = WiggleReader(reference, start, end)
        try:
            # run command and grab output simultaneously
            process = subprocess.Popen(cmd, stdout=subprocess.PIPE)
            while True:
                line = process.stdout.readline()
                if line == '' and process.poll() is not None:
                    break
                wiggleReader.readWiggleLine(line.strip())
        except ValueError:
            raise
        except:
            raise Exception("bigWigToWig failed to run")

        return wiggleReader.getData()

    def bigWigToProtocol(self, reference, start, end):
        # return self.readValuesBigWigToWig(reference, start, end)
        for continuousObj in self.readValuesPyBigWig(reference, start, end):
            yield continuousObj


class AbstractContinuousSet(datamodel.DatamodelObject):
    """
    A continuous sequence annotation set
    """
    compoundIdClass = datamodel.ContinuousSetCompoundId

    def __init__(self, parentContainer, localId):
        super(AbstractContinuousSet, self).__init__(parentContainer, localId)
        self._name = localId
        self._sourceUri = ""
        self._referenceSet = None

    def getReferenceSet(self):
        """
        Returns the reference set associated with this ContinuousSet.
        """
        return self._referenceSet

    def setReferenceSet(self, referenceSet):
        """
        Sets the reference set associated with this ContinuousSet to the
        specified value.
        """
        self._referenceSet = referenceSet

    def toProtocolElement(self):
        """
        Returns the representation of this ContinuousSet as the corresponding
        ProtocolElement.
        """
        gaContinuousSet = protocol.ContinuousSet()
        gaContinuousSet.id = self.getId()
        gaContinuousSet.dataset_id = self.getParentContainer().getId()
        gaContinuousSet.reference_set_id = pb.string(
                                            self._referenceSet.getId())
        gaContinuousSet.name = self._name
        gaContinuousSet.source_uri = self._sourceUri
        attributes = self.getAttributes()
        for key in attributes:
            gaContinuousSet.attributes.attr[key] \
                .values.extend(protocol.encodeValue(attributes[key]))
        return gaContinuousSet


class FileContinuousSet(AbstractContinuousSet):
    """
    Data associated with a file containing continuous data.
    """
    def __init__(self, parentContainer, localId):
        super(FileContinuousSet, self).__init__(parentContainer, localId)
        self._filePath = None

    def populateFromFile(self, dataUrl):
        """
        Populates the instance variables of this ContinuousSet from the
        specified data URL.
        """
        self._filePath = dataUrl

    def populateFromRow(self, continuousSetRecord):
        """
        Populates the instance variables of this ContinuousSet from the
        specified DB row.
        """
        self._filePath = continuousSetRecord.dataurl
        self.setAttributesJson(continuousSetRecord.attributes)

    def getDataUrl(self):
        """
        Returns the URL providing the data source for this ContinuousSet.
        """
        return self._filePath

    def getContinuous(self, referenceName=None, start=None, end=None):
        """
        Method passed to runSearchRequest to fulfill the request to
        yield continuous protocol objects that satisfy the given query.

        :param str referenceName: name of reference (ex: "chr1")
        :param start: castable to int, start position on reference
        :param end: castable to int, end position on reference
        :return: yields a protocol.Continuous at a time
        """
        bigWigReader = BigWigDataSource(self._filePath)
        for continuousObj in bigWigReader.bigWigToProtocol(
                                            referenceName, start, end):
            yield continuousObj


class SimulatedContinuousSet(AbstractContinuousSet):
    """
    Simulated data backend for ContinuousSet, used for internal testing.
    """
    def __init__(self, parentContainer, localId, randomSeed=1):
        self._randomSeed = randomSeed
        super(SimulatedContinuousSet, self).__init__(parentContainer, localId)

    def _generateSimulatedContinuous(self, randomNumberGenerator):
        continuous = protocol.Continuous()
        continuous.start = randomNumberGenerator.randint(1000, 2000)
        continuous.values = [100, 200.3, 400]

    def getContinuousData(self, referenceName=None, start=None, end=None):
        """
        Returns a set number of simulated continuous data.

        :param referenceName: name of reference to "search" on
        :param start: start coordinate of query
        :param end: end coordinate of query
        :return: Yields continuous list
        """
        randomNumberGenerator = random.Random()
        randomNumberGenerator.seed(self._randomSeed)
        for i in range(100):
            gaContinuous = self._generateSimulatedContinuous(
                                    randomNumberGenerator)
            match = (
                gaContinuous.start < end and
                gaContinuous.end > start and
                gaContinuous.reference_name == referenceName)
            if match:
                yield gaContinuous



"""
Unit tests for the peer data model.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

from ga4gh.server.datamodel import peers
from ga4gh.server import exceptions


class TestPeers(unittest.TestCase):
    """
    Unit tests for the peer datamodel.
    """
    def testBadUrlException(self):
        peers.Peer("https://1kgenomes.ga4gh.org")
        with (self.assertRaises(exceptions.BadUrlException)):
            peers.Peer("ht9://1kgenomes.ga4gh.org")
        with (self.assertRaises(exceptions.BadUrlException)):
            peers.Peer("http://1kgen")
        with (self.assertRaises(exceptions.BadUrlException)):
            peers.Peer("http://.org")

    def testToProtocolElement(self):
        url = "http://1kgenomes.ga4gh.org"
        key = "testkey"
        value = "testvalue"
        attributes = {key: value}
        peer = peers.Peer(url, attributes)
        protocolElement = peer.toProtocolElement()
        self.assertEqual(url, protocolElement.url)
        self.assertEqual(
            attributes[key],
            protocolElement.attributes.attr[key].values[0].string_value)

    def testAttributesJson(self):
        url = "http://1kgenomes.ga4gh.org"
        with (self.assertRaises(exceptions.InvalidJsonException)):
            peer = peers.Peer(url)
            peer.setAttributesJson('a{"bad": "json"}')
        with (self.assertRaises(exceptions.InvalidJsonException)):
            peer = peers.Peer(url)
            peer.setAttributesJson('{"bad"; "json"}')
        with (self.assertRaises(exceptions.InvalidJsonException)):
            peer = peers.Peer(url)
            peer.setAttributesJson('{"bad": json}')



"""
Server cli
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import requests

import ga4gh.server.cli as cli
import ga4gh.server.frontend as frontend

import ga4gh.common.cli as common_cli


def addServerOptions(parser):
    parser.add_argument(
        "--port", "-P", default=8000, type=int,
        help="The port to listen on")
    parser.add_argument(
        "--host", "-H", default="127.0.0.1",
        help="The server host string; use 0.0.0.0 to allow all connections.")
    parser.add_argument(
        "--config", "-c", default='DevelopmentConfig', type=str,
        help="The configuration to use")
    parser.add_argument(
        "--config-file", "-f", type=str, default=None,
        help="The configuration file to use")
    parser.add_argument(
        "--tls", "-t", action="store_true", default=False,
        help="Start in TLS (https) mode.")
    parser.add_argument(
        "--dont-use-reloader", default=False, action="store_true",
        help="Don't use the flask reloader")
    cli.addVersionArgument(parser)
    cli.addDisableUrllibWarningsArgument(parser)


def getServerParser():
    parser = common_cli.createArgumentParser("GA4GH reference server")
    addServerOptions(parser)
    return parser


def server_main(args=None):
    parser = getServerParser()
    parsedArgs = parser.parse_args(args)
    if parsedArgs.disable_urllib_warnings:
        requests.packages.urllib3.disable_warnings()
    frontend.configure(
        parsedArgs.config_file, parsedArgs.config, parsedArgs.port)
    sslContext = None
    if parsedArgs.tls or ("OIDC_PROVIDER" in frontend.app.config):
        sslContext = "adhoc"
    frontend.app.run(
        host=parsedArgs.host, port=parsedArgs.port,
        use_reloader=not parsedArgs.dont_use_reloader,
        ssl_context=sslContext)



"""
Tests the datasets module
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.server.datamodel.datasets as datasets


class TestDatasets(unittest.TestCase):
    """
    Tests the datasets class
    """
    def testToProtocolElement(self):
        datasetId = 'ds1'
        dataset = datasets.SimulatedDataset(datasetId, 1, 2, 3, 4, 5)
        dataset.setAttributes({"test": "test"})
        gaDataset = dataset.toProtocolElement()
        self.assertIsNotNone(gaDataset.attributes.attr)
        self.assertEqual(
            gaDataset.attributes.attr['test'].values[0].string_value, "test")
        self.assertEqual(dataset.getId(), gaDataset.id)
